#ifndef KRY_NETWORK_PACKET_H
#define KRY_NETWORK_PACKET_H

#include "Utilities\StringTypes.h"
#include <vector>

namespace kry
{
	namespace Network
	{
		class Packet;

		template <typename Type, unsigned typesize = sizeof(Type)>
		Packet& operator<<(Packet& left, const Type& right);
		template <typename Type, unsigned typesize = sizeof(Type)>
		Packet& operator>>(Packet& left, Type& right);

		class Packet
		{
			public:
				Packet();
				Packet(unsigned size);
				Packet(const Packet& packet);
				Packet(Packet&& packet);
				virtual ~Packet();

				template <typename Type>
				Packet& putBytes(const Type& bytes, unsigned length = sizeof(Type));
				template<> Packet& putBytes<const Util::String&>(const Util::String& bytes, unsigned length);
				template<> Packet& putBytes<const Packet&>(const Packet& bytes, unsigned length);

				template <typename Type>
				Packet& getBytes(Type& bytes, unsigned length = sizeof(Type));
				template<> Packet& getBytes<Util::String&>(Util::String& bytes, unsigned length);
				template<> Packet& getBytes<Packet&>(Packet& bytes, unsigned length);

				inline void resize(unsigned size);
				inline Packet& setIterator(unsigned position);
				inline unsigned getIterator() const;
				inline char& at(unsigned index);
				inline const char& at(unsigned index) const;
				inline char& operator[](unsigned index);
				inline const char& operator[](unsigned index) const;
				inline char* getData();
				inline const char* getData() const;
				inline unsigned getSize() const;
				static inline unsigned getHeaderSize();

				virtual inline Packet& operator=(const Packet& packet);
				virtual inline Packet& operator=(Packet&& packet);

			protected:
				typedef unsigned HeaderType;

				inline bool checkIndex(unsigned index) const;
				inline void updateHeader(unsigned size);
				
				static const unsigned char headersize = sizeof(HeaderType);
				std::vector<char> data;
				const unsigned limit = data.max_size() - headersize;
				unsigned itr;
		};
	}
}

#include "Utilities\Exception.h"
#include "Utilities\BasicString.h"

namespace kry
{
	namespace Network
	{
		template <typename Type>
		Packet& Packet::putBytes(const Type& bytes, unsigned length)
		{
			if (static_cast<unsigned long long>(itr) + length > limit)
				throw Util::Exception("Data size exceeds maximum limit! [length=" + Util::toString(length) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (static_cast<unsigned long long>(itr) + length > data.size())
				throw Util::Exception("Data size exceeds packet size! [length=" + Util::toString(length) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			const char* b = reinterpret_cast<const char*>(&bytes);
			data.reserve(data.size() + length);
			for (unsigned i = 0; i < length; ++i, ++b, ++itr)
				data.at(itr) = *b;

			return *this;
		}

		template<>
		Packet& Packet::putBytes<const Util::String&>(const Util::String& bytes, unsigned length)
		{
			length = bytes.getLength();
			if (static_cast<unsigned long long>(itr) + length > limit)
				throw Util::Exception("Data size exceeds maximum limit! [length=" + Util::toString(length) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (static_cast<unsigned long long>(itr) + length > data.size())
				throw Util::Exception("Data size exceeds packet size! [length=" + Util::toString(length) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			data.reserve(data.size() + length);
			for (unsigned i = 0; i < length; ++i, ++itr)
				data.at(itr) = bytes[i];

			return *this;
		}

		template<>
		Packet& Packet::putBytes<const Packet&>(const Packet& bytes, unsigned length)
		{
			length = bytes.getSize();
			if (static_cast<unsigned long long>(itr) + length > limit)
				throw Util::Exception("Data size exceeds maximum limit! [length=" + Util::toString(length) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (static_cast<unsigned long long>(itr) + length > data.size())
				throw Util::Exception("Data size exceeds packet size! [length=" + Util::toString(length) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			data.reserve(data.size() + length);
			for (unsigned i = 0; i < length; ++i, ++itr)
				data.at(itr) = bytes[i];

			return *this;
		}

		template <typename Type>
		Packet& Packet::getBytes(Type& bytes, unsigned length)
		{
			if (static_cast<unsigned long long>(itr) + length > limit)
				throw Util::Exception("Data size exceeds maximum limit! [length=" + Util::toString(length) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (static_cast<unsigned long long>(itr) + length > data.size())
				throw Util::Exception("Data size exceeds packet size! [length=" + Util::toString(length) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			char* b = reinterpret_cast<char*>(&bytes);
			for (unsigned i = 0; i < length; ++i, ++b, ++itr)
				*b = data.at(itr);

			return *this;
		}

		template<>
		Packet& Packet::getBytes<Util::String&>(Util::String& bytes, unsigned length)
		{
			length = bytes.getLength();
			if (static_cast<unsigned long long>(itr) + length > limit)
				throw Util::Exception("Data size exceeds maximum limit! [length=" + Util::toString(length) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (static_cast<unsigned long long>(itr) + length > data.size())
				throw Util::Exception("Data size exceeds packet size! [length=" + Util::toString(length) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			for (unsigned i = 0; i < length; ++i, ++itr)
				bytes[i] = data.at(itr);

			return *this;
		}

		template<>
		Packet& Packet::getBytes<Packet&>(Packet& bytes, unsigned length)
		{
			length = bytes.getSize();
			if (static_cast<unsigned long long>(itr) + length > limit)
				throw Util::Exception("Data size exceeds maximum limit! [length=" + Util::toString(length) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (static_cast<unsigned long long>(itr) + length > data.size())
				throw Util::Exception("Data size exceeds packet size! [length=" + Util::toString(length) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			for (unsigned i = 0; i < length; ++i, ++itr)
				bytes.at(i) = data.at(itr);

			return *this;
		}

		void Packet::resize(unsigned size)
		{
			if (size >= limit)
				throw Util::Exception("Size greater than maximum limit! [size=" + Util::toString(size) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			data.resize(size + headersize);
			updateHeader(data.size());
		}

		Packet& Packet::setIterator(unsigned position)
		{
			if (position >= limit)
				throw Util::Exception("Position greater than maximum limit! [position=" + Util::toString(position) + ", limit=" + Util::toString(limit) + ']', KRY_EXCEPTION_DATA);
			if (position >= data.size())
				throw Util::Exception("Position greater than packet size! [position=" + Util::toString(position) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);

			itr = position + headersize;
			return *this;
		}

		unsigned Packet::getIterator() const
		{
			return itr - headersize;
		}

		char& Packet::at(unsigned index)
		{
			if (!checkIndex(index))
				throw Util::Exception("Index out of bounds! [index=" + Util::toString(index) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);
			return data.at(index + headersize);
		}

		const char& Packet::at(unsigned index) const
		{
			if (!checkIndex(index))
				throw Util::Exception("Index out of bounds! [index=" + Util::toString(index) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);
			return data.at(index + headersize);
		}

		char& Packet::operator[](unsigned index)
		{
			if (!checkIndex(index))
				throw Util::Exception("Index out of bounds! [index=" + Util::toString(index) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);
			return data[index + headersize];
		}

		const char& Packet::operator[](unsigned index) const
		{
			if (!checkIndex(index))
				throw Util::Exception("Index out of bounds! [index=" + Util::toString(index) + ", size=" + Util::toString(data.size()) + ']', KRY_EXCEPTION_DATA);
			return data[index + headersize];
		}

		char* Packet::getData()
		{
			return data.data();
		}

		const char* Packet::getData() const
		{
			return data.data();
		}

		unsigned Packet::getSize() const
		{
			return data.size() - headersize;
		}

		unsigned Packet::getHeaderSize()
		{
			return headersize;
		}

		bool Packet::checkIndex(unsigned index) const
		{
			return index < data.size() && index < limit;
		}

		void Packet::updateHeader(unsigned size)
		{
			auto head = reinterpret_cast<HeaderType*>(data.data());
			*head = static_cast<HeaderType>(size);
		}

		Packet& Packet::operator=(const Packet& packet)
		{
			data = packet.data;
			itr = packet.itr;
			updateHeader(data.size());
			return *this;
		}

		Packet& Packet::operator=(Packet&& packet)
		{
			data = std::move(packet.data);
			itr = packet.itr;
			updateHeader(data.size());
			return *this;
		}


		template <typename Type, unsigned typesize>
		Packet& operator<<(Packet& left, const Type& right)
		{
			return left.putBytes(right, typesize);
		}

		template <typename Type, unsigned typesize>
		Packet& operator>>(Packet& left, Type& right)
		{
			return left.getBytes(right, typesize);
		}
	}
}

#endif
