#ifndef KRY_UTILITIES_BASICVECTOR_H
#define KRY_UTILITIES_BASICVECTOR_H

#include <iostream>
#include <type_traits>
#include "Utilities\String.h"
#include "Utilities\StringConvert.h"

namespace kry
{
	namespace Util
	{
		template<typename VectorType, unsigned int VectorSize>
		struct Vector
		{
			VectorType elem[VectorSize];

			// Construction
			Vector(const VectorType Default = VectorType())
			{
				fill(Default);
			}

			Vector(const VectorType A, const VectorType B)
			{
				fill(VectorType());

				elem[0] = A;
				elem[1] = B;
			}

			Vector(const VectorType A, const VectorType B, const VectorType C)
			{
				fill(VectorType());

				elem[0] = A;
				elem[1] = B;
				if (VectorSize < 3) return;
				elem[2] = C;
			}

			Vector(const VectorType A, const VectorType B, const VectorType C, const VectorType D)
			{
				fill(VectorType());

				elem[0] = A;
				elem[1] = B;
				if (VectorSize < 3) return;
				elem[2] = C;
				if (VectorSize < 4) return;
				elem[3] = D;
			}

			template<typename OtherVectorType, unsigned int OtherVectorSize>
			Vector(const Vector<OtherVectorType, OtherVectorSize>& From)
			{
				fill(VectorType());

				for (unsigned int i = 0; i < VectorSize && i < OtherVectorSize; ++i)
					elem[i] = static_cast<OtherVectorType>(From[i]);
			}

			Vector(const String& From)
			{
				fill(VectorType());

				if (std::is_arithmetic<VectorType>::value)
				{
					auto explosion = From.explode(',');

					for (unsigned int i = 0; i < VectorSize && i < explosion.size(); ++i)
					{
						float value = 0.0f;

						try
						{
							value = toDecimal<float>(trimNumber(explosion[i]));
						}
						catch (std::exception e)
						{
							throw Exception("Could not convert String to Vector, invalid number value! [String=" + From + "][value=" + trimNumber(explosion[i]) + ']', KRY_EXCEPTION_DATA);
						}

						elem[i] = static_cast<VectorType>(value);
					}
				}
			}

			// Copy
			Vector(const Vector& From)
			{
				if (&From != this)
				{
					for (unsigned int i = 0; i < VectorSize; ++i)
						elem[i] = From[i];
				}
			}

			void operator=(const Vector& From)
			{
				if (&From != this)
				{
					for (unsigned int i = 0; i < VectorSize; ++i)
						elem[i] = From[i];
				}
			}

			// Access
			VectorType& operator[](unsigned int Pos)
			{
				return elem[Pos];
			}

			VectorType operator[](unsigned int Pos) const
			{
				return elem[Pos];
			}

			// Math operators
			void operator+=(const Vector& From)
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
					elem[i] += From[i];
			}

			void operator-=(const Vector& From)
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
					elem[i] -= From[i];
			}

			void operator*=(const Vector& From)
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
					elem[i] *= From[i];
			}

			void operator/=(const Vector& From)
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
					elem[i] /= From[i];
			}

			Vector operator+(const Vector& From) const
			{
				Vector temp = *this;
				temp += From;

				return temp;
			}

			Vector operator-(const Vector& From) const
			{
				Vector temp = *this;
				temp -= From;

				return temp;
			}

			Vector operator*(const Vector& From) const
			{
				Vector temp = *this;
				temp *= From;

				return temp;
			}

			Vector operator/(const Vector& From) const
			{
				Vector temp = *this;
				temp /= From;

				return temp;
			}

			// Comparison operators
			bool operator==(const Vector& From) const
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
				{
					if (elem[i] != From[i])
						return false;
				}

				return true;
			}

			bool operator!=(const Vector& From) const
			{
				return !(*this == From);
			}

			bool operator<(const Vector& From) const
			{
				for (int i = int(VectorSize) - 1; i >= 0; --i)
				{
					if (elem[i] < From.elem[i])
						return true;

					if (elem[i] != From.elem[i])
						return false;
				}

				return false;
			}

			bool operator>(const Vector& From) const
			{
				return From.operator<(*this);
			}

			bool operator<=(const Vector& From) const
			{
				return !From.operator<(*this);
			}

			bool operator>=(const Vector& From) const
			{
				return !operator<(From);
			}

			// Operations
			Vector zero()
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
					elem[i] = 0;

				return *this;
			}

			Vector fill(const VectorType Value)
			{
				for (unsigned int i = 0; i < VectorSize; ++i)
					elem[i] = Value;

				return *this;
			}

			// Information
			unsigned int size() const
			{
				return VectorSize;
			}

			// Overloads
			friend std::ostream & operator<<(std::ostream &Stream, const Vector& From)
			{
				Stream << "{";

				for (unsigned int i = 0; i < From.size() - 1; ++i)
					Stream << From[i] << ", ";

				Stream << From[From.size() - 1] << "}";

				return Stream;
			}

			// Converters
			String toString() const
			{
				String temp = "{";

				for (unsigned int i = 0; i < VectorSize - 1; ++i)
					temp += Util::toString(elem[i]) + ", ";

				temp += Util::toString(elem[VectorSize - 1]) + "}";

				return temp;
			}

			private:
				String trimNumber(const String& Dirty)
				{
					unsigned int left = 0;
					unsigned int right = Dirty.getLength();

					if (right > 0)
					{
						for (unsigned int i = 0; i < Dirty.getLength(); ++i)
						{
							if ((Dirty[i] >= '0' && Dirty[i] <= '9') || Dirty[i] == '.' || Dirty[i] == '-')
							{
								left = i;
								break;
							}
						}

						for (unsigned int i = left + 1; i < Dirty.getLength(); ++i)
						{
							if (Dirty[i] < '0' && Dirty[i] > '9' && Dirty[i] != '.' && Dirty[i] != '-')
							{
								right = i;
								break;
							}
						}

						return String(&Dirty[left], right - left);
					}

					return Dirty;
				}
		};
	}
}

#endif
