#ifndef KRY_UTILITIES_STRINGCONVERT_H
#define KRY_UTILITIES_STRINGCONVERT_H

#include "Utilities\Exception.h"
#include <type_traits>

namespace kry
{
	namespace Util
	{
		enum class IntegralBase
		{
			BINARY,
			DECIMAL,
			OCTAL,
			HEXADECIMAL
		};

		template <typename T, typename = std::enable_if<std::is_arithmetic<T>::value>::type>
		String toString(T value, IntegralBase base = IntegralBase::DECIMAL);
		template <typename T>
		String toString(T value, IntegralBase base, std::true_type isintegral);
		template <typename T>
		String toString(T value, IntegralBase base, std::false_type isfloating);

		template <typename T, typename = std::enable_if<std::is_integral<T>::value>::type>
		T toIntegral(const String& string);
		template <typename T, typename Unsigned = std::make_unsigned<T>::type, typename = std::enable_if<std::is_integral<T>::value>::type>
		Unsigned toUIntegral(const String& string);
		template <typename T, typename = std::enable_if<std::is_floating_point<T>::value>::type>
		T toDecimal(const String& string);
	}
}

#include "BasicString.h"
#include <..\FMT\format.h>

namespace kry
{
	namespace Util
	{
		template <typename T, typename>
		String toString(T value, IntegralBase base)
		{
			return toString(value, base, std::is_integral<T>());
		}

		template <typename T>
        String toString(T value, IntegralBase base, std::true_type)
		{
			switch (base)
			{
				case IntegralBase::BINARY:
					{
						std::string ret = std::move(fmt::format("{0:b}", value));
						return String(ret.data(), ret.length());
					}
				case IntegralBase::DECIMAL:
					{
						fmt::FormatInt format(value);
						return String(format.data(), format.size());
					}
				case IntegralBase::OCTAL:
					{
						std::string ret = std::move(fmt::format("{0:o}", value));
						return String(ret.data(), ret.length());
					}
				case IntegralBase::HEXADECIMAL:
					{
						std::string ret = std::move(fmt::format("{0:x}", value));
						return String(ret.data(), ret.length());
					}
			}
			throw Exception("Failed to convert integral to string! Invalid base.", KRY_EXCEPTION_DATA);
		}

		template <typename T>
        String toString(T value, IntegralBase, std::false_type)
		{
			std::string ret = std::move(fmt::format("{:g}", value));
			return String(ret.data(), ret.length());
		}

		template <typename T, typename>
		T toIntegral(const String& string)
		{
			try
			{
				std::string convert(string.getData(), string.getLength());
				return static_cast<T>(std::stoll(convert));
			}
			catch (const std::exception& e)
			{
				throw Exception("Failed to convert string to integral! Invalid argument. [string=" + string + ']', KRY_EXCEPTION_DATA);
			}
		}

		template <typename T, typename Unsigned, typename>
		Unsigned toUIntegral(const String& string)
		{
			try
			{
				std::string convert(string.getData(), string.getLength());
				return static_cast<Unsigned>(std::stoull(convert));
			}
			catch (const std::exception& e)
			{
				throw Exception("Failed to convert string to unsigned integral! Invalid argument. [string=" + string + ']', KRY_EXCEPTION_DATA);
			}
		}

		template <typename T, typename>
		T toDecimal(const String& string)
		{
			try
			{
				std::string convert(string.getData(), string.getLength());
				return static_cast<T>(std::stold(convert));
			}
			catch (const std::exception& e)
			{
				throw Exception("Failed to convert string to decimal! Invalid argument. [string=" + string + ']', KRY_EXCEPTION_DATA);
			}
		}
	}
}

#endif
