#include "Audio\Buffer.h"
#include "Utilities\String.h"
#include "OpenAL\al.h"
#include "OpenAL\alc.h"

#define KRY_INVALID_BUFFER 0

namespace kry
{
	namespace Audio
	{
		Util::String formatToString(Format format)
		{
			switch (format)
			{
				case Format::MONO_8: return Util::String("MONO_8");
				case Format::MONO_16: return Util::String("MONO_16");
				case Format::STEREO_8: return Util::String("STEREO_8");
				case Format::STEREO_16: return Util::String("STEREO_16");
				case Format::QUAD_16: return Util::String("QUAD_16");
				case Format::CHANNEL_51_16: return Util::String("CHANNEL_51_16");
				case Format::CHANNEL_61_16: return Util::String("CHANNEL_61_16");
				case Format::CHANNEL_71_16: return Util::String("CHANNEL_71_16");
			}
			return Util::String("UNKNOWN");
		}

		Buffer::Buffer() : id(KRY_INVALID_BUFFER)
		{
			alGenBuffers(1, &id);
			if (id == KRY_INVALID_BUFFER)
				throw Util::Exception("Failed to create buffer! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		Buffer::Buffer(const std::vector<unsigned char>& samples, Format format, unsigned frequency) : id(KRY_INVALID_BUFFER)
		{
			alGenBuffers(1, &id);
			if (id == KRY_INVALID_BUFFER)
				throw Util::Exception("Failed to create buffer! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			ALenum f = formatToALFormat(format);
			if (f == -1)
				throw Util::Exception("Failed to create buffer! Unknown format. [format=" + formatToString(format) + ']', KRY_EXCEPTION_DATA);
			alBufferData(id, f, reinterpret_cast<const void*>(samples.data()), samples.size(), frequency);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set buffer data! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		Buffer::Buffer(const unsigned char* samples, unsigned size, Format format, unsigned frequency) : id(KRY_INVALID_BUFFER)
		{
			alGenBuffers(1, &id);
			if (id == KRY_INVALID_BUFFER)
				throw Util::Exception("Failed to create buffer! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			ALenum f = formatToALFormat(format);
			if (f == -1)
				throw Util::Exception("Failed to create buffer! Unknown format. [format=" + formatToString(format) + ']', KRY_EXCEPTION_DATA);
			alBufferData(id, f, reinterpret_cast<const void*>(samples), size, frequency);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set buffer data! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		Buffer::~Buffer()
		{
			if (id != KRY_INVALID_BUFFER)
				alDeleteBuffers(1, &id);
		}

		void Buffer::resetData(const std::vector<unsigned char>& samples, Format format, unsigned frequency)
		{
			if (id == KRY_INVALID_BUFFER)
				throw Util::Exception("Buffer is invalid!", KRY_EXCEPTION_DATA);

			int f = formatToALFormat(format);
			if (f == -1)
				throw Util::Exception("Failed to reset buffer! Unknown format. [format=" + formatToString(format) + ']', KRY_EXCEPTION_DATA);
			alBufferData(id, f, reinterpret_cast<const void*>(samples.data()), samples.size(), frequency);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to reset buffer data! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		ALenum Buffer::formatToALFormat(Format format)
		{
			switch (format)
			{
				case Format::MONO_8: return AL_FORMAT_MONO8;
				case Format::MONO_16: return AL_FORMAT_MONO16;
				case Format::STEREO_8: return AL_FORMAT_STEREO8;
				case Format::STEREO_16: return AL_FORMAT_STEREO16;
				case Format::QUAD_16: return alGetEnumValue("AL_FORMAT_QUAD16");
				case Format::CHANNEL_51_16: return alGetEnumValue("AL_FORMAT_51CHN16");
				case Format::CHANNEL_61_16: return alGetEnumValue("AL_FORMAT_61CHN16");
				case Format::CHANNEL_71_16: return alGetEnumValue("AL_FORMAT_71CHN16");
			}
			return -1;
		}
	}
}
