#include "Audio\Source.h"
#include "Audio\Buffer.h"
#include "Utilities\Exception.h"
#include "Utilities\BasicString.h"
#include "OpenAL\al.h"
#include <algorithm>
#include <Windows.h>

#define KRY_INVALID_SOURCE 0

namespace kry
{
	namespace Audio
	{
		Source::Source() : id(KRY_INVALID_SOURCE), volume(1.0f), pitch(1.0f), distance(1.0f), attenuation(1.0f), looping(false), binds(0)
		{
			alGenSources(1, &id);
			if (id == KRY_INVALID_SOURCE)
				throw Util::Exception("Failed to create source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			alSourcei(id, AL_BUFFER, AL_NONE);

			setLooping(looping);
		}

		Source::Source(const Buffer& buffer) : id(KRY_INVALID_SOURCE), volume(1.0f), pitch(1.0f), distance(1.0f), attenuation(1.0f), looping(false), binds(0)
		{
			bufferqueue.push_back(const_cast<Buffer*>(&buffer));
			alGenSources(1, &id);
			if (id == KRY_INVALID_SOURCE)
				throw Util::Exception("Failed to create source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			alSourcei(id, AL_BUFFER, buffer.getID());
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to attach buffer to source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);

			setLooping(looping);
		}

		Source::Source(const Source& copy) : bufferqueue(copy.bufferqueue)
		{
			alGenSources(1, &id);
			if (id == KRY_INVALID_SOURCE)
				throw Util::Exception("Failed to create source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			if (!bufferqueue.empty())
				alSourcei(id, AL_BUFFER, (*bufferqueue.begin())->getID());
			else
				alSourcei(id, AL_BUFFER, AL_NONE);

			setPosition(copy.getPosition());
			setVelocity(copy.getVelocity());
			setVolume(copy.getVolume());
			setPitch(copy.getPitch());
			setDistance(copy.getDistance());
			setAttenuation(copy.getAttenuation());
			setLooping(copy.isLooping());
			binds = copy.binds;
		}

		Source::Source(Source&& move) : bufferqueue(std::move(move.bufferqueue)), position(move.position), velocity(move.velocity), volume(move.volume), pitch(move.pitch), 
			distance(move.distance), attenuation(move.attenuation), id(move.id), looping(move.looping), binds(move.binds)
		{
			move.id = KRY_INVALID_SOURCE;
		}

		Source::~Source()
		{
			destroy();
		}

		void Source::setPosition(const Util::Vector3f& newposition)
		{
			position = newposition;
			alSourcefv(id, AL_POSITION, position.elem);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set position! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setVelocity(const Util::Vector3f& newvelocity)
		{
			velocity = newvelocity;
			alSourcefv(id, AL_VELOCITY, velocity.elem);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set velocity! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setVolume(float newvolume)
		{
			volume = newvolume;
			alSourcef(id, AL_GAIN, volume);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set volume! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setPitch(float newpitch)
		{
			pitch = newpitch;
			alSourcef(id, AL_PITCH, pitch);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set pitch! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setDistance(float newdistance)
		{
			distance = newdistance;
			alSourcef(id, AL_REFERENCE_DISTANCE, distance);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set distance! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setAttenuation(float newattenuation)
		{
			attenuation = newattenuation;
			alSourcef(id, AL_ROLLOFF_FACTOR, attenuation);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set attenuation! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setNewBuffer(const Buffer& buffer)
		{
			alSourceStop(id);
			bufferqueue.clear();
			bufferqueue.push_back(const_cast<Buffer*>(&buffer));
			alSourcei(id, AL_BUFFER, buffer.getID());
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to attach new buffer to source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setLooping(bool loop)
		{
			looping = loop;
			alSourcei(id, AL_LOOPING, looping ? AL_TRUE : AL_FALSE);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to toggle looping! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::setOffset(float secondsoffset)
		{
			alSourcef(id, AL_SEC_OFFSET, secondsoffset);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to set time offset! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		float Source::getOffset() const
		{
			float seconds;
			alGetSourcef(id, AL_SEC_OFFSET, &seconds);
			return seconds;
		}

		void Source::play()
		{
			alSourcePlay(id);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to play source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::pause()
		{
			alSourcePause(id);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to pause source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::stop()
		{
			alSourceStop(id);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to stop source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
		}

		void Source::queue(const Buffer& buffer)
		{
			bufferqueue.push_back(const_cast<Buffer*>(&buffer));
			unsigned buf = buffer.getID();
			alSourceQueueBuffers(id, 1, &buf);
		}

		void Source::unqueue(const Buffer& buffer)
		{
			unsigned buf = buffer.getID();
			alSourceUnqueueBuffers(id, 1, &buf);
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to un-queue buffer! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			auto itr = std::find(bufferqueue.begin(), bufferqueue.end(), &buffer);
			bufferqueue.erase(itr);
		}

		void Source::clearQueue()
		{
			bufferqueue.clear();
			alSourceStop(id);
			alSourcei(id, AL_BUFFER, AL_NONE);
		}

		Status Source::getStatus() const
		{
			int status;
			alGetSourcei(id, AL_SOURCE_STATE, &status);

			switch (status)
			{
				case AL_PAUSED: return Status::PAUSED;
				case AL_PLAYING: return Status::PLAYING;
			}

			return Status::STOPPED;
		}

		const Buffer* Source::getCurrBuffer() const /** #TODO(test) not sure if getsource will return head of queue or current queue */
		{
			unsigned cur;
			alGetSourcei(id, AL_BUFFER, reinterpret_cast<int*>(&cur));
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to get current buffer! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);

			auto itr = std::find_if(bufferqueue.begin(), bufferqueue.end(), [cur](Buffer* buffer) -> bool 
			{
				return buffer->getID() == cur;
			});

			return itr == bufferqueue.end() ? nullptr : *itr;
		}

		unsigned Source::getCurrBufferIndex() const /** #TODO(test) not sure if getsource will return head of queue or current queue */
		{
			unsigned cur;
			alGetSourcei(id, AL_BUFFER, reinterpret_cast<int*>(&cur));
			if (alGetError() != AL_NO_ERROR)
				throw Util::Exception("Failed to get current buffer! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);

			unsigned index = 0; // (unsigned equivalent of -1, wrap around to 0 upon increment)
			auto end = bufferqueue.end();
			for (auto itr = bufferqueue.begin(); itr != end; ++itr, ++index)
				if ((*itr)->getID() == cur)
					return index;
			return 0xfffffffe;
		}

		Source& Source::operator=(const Source& copy)
		{
			destroy();

			bufferqueue = copy.bufferqueue;
			alGenSources(1, &id);
			if (id == KRY_INVALID_SOURCE)
				throw Util::Exception("Failed to create source! [ALerrorcode=" + Util::toString(alGetError()) + ']', KRY_EXCEPTION_DATA);
			if (!bufferqueue.empty())
				alSourcei(id, AL_BUFFER, (*bufferqueue.begin())->getID());
			else
				alSourcei(id, AL_BUFFER, AL_NONE);

			setPosition(copy.getPosition());
			setVelocity(copy.getVelocity());
			setVolume(copy.getVolume());
			setPitch(copy.getPitch());
			setDistance(copy.getDistance());
			setAttenuation(copy.getAttenuation());
			setLooping(copy.isLooping());

			return *this;
		}

		Source& Source::operator=(Source&& move)
		{
			destroy();
			
			bufferqueue = std::move(move.bufferqueue);
			position = move.position;
			velocity = move.velocity;
			volume = move.volume;
			pitch = move.pitch;
			distance = move.distance;
			attenuation = move.attenuation;
			id = move.id;
			looping = move.looping;

			move.id = KRY_INVALID_SOURCE;

			return *this;
		}

		void Source::destroy()
		{
			if (id != KRY_INVALID_SOURCE)
			{
				stop();

				while (binds > 0)
					Sleep(1);

				alSourcei(id, AL_BUFFER, AL_NONE);
				alDeleteSources(1, &id);
				id = KRY_INVALID_SOURCE;
			}
		}
	}
}
