#include "Graphics\Canvas.h"
#include "Graphics\Texture.h"
#include "Graphics\Font.h"
#include "Utilities\String.h"
#include "Utilities\Exception.h"
#include "System\Window.h"
#include <Windows.h>
#undef min
#undef max
#include <algorithm>
#include <GL\glew.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <ft2build.h>
#include FT_FREETYPE_H

namespace kry
{
	namespace Graphics
	{
		Canvas::Canvas()
		{
			fboID = 0;
			colorID = 0;
			scale.fill(1);
			rgba.fill(1);
			rotation = 0;
			linearFilter = true;
			batchCapacityCache = 0;
		}

		Canvas::~Canvas()
		{
			deleteBuffers();
		}

		void Canvas::clear()
		{
			batchCapacityCache = typeBatch.size();
			typeBatch.clear();
			graphicBatch.clear();

			for (auto i = glyphs.begin(); i != glyphs.end(); ++i)
			{
				for (auto j = i->second.begin(); j != i->second.end(); ++j)
					delete *j;
			}

			glyphs.clear();
			typeBatch.reserve(batchCapacityCache);
			graphicBatch.reserve(batchCapacityCache);
		}

		void Canvas::deleteBuffers()
		{
			if (fboID)
				glDeleteFramebuffersEXT(1, &fboID);

			if (colorID)
				glDeleteTextures(1, &colorID);

			fboID = 0;
			colorID = 0;
		}

		void Canvas::addSprite(Sprite* sprite)
		{
			if (sprite != nullptr)
			{
				typeBatch.push_back(GraphicObjectType::SPRITE);
				graphicBatch.push_back(static_cast<void*>(sprite));
			}		
		}

		void Canvas::addSprites(const std::vector<Sprite*>& sprites)
		{
			if (!sprites.empty())
			{
				typeBatch.insert(typeBatch.end(), sprites.size(), GraphicObjectType::SPRITE);
				graphicBatch.insert(graphicBatch.end(), sprites.begin(), sprites.end());
			}
		}

		void Canvas::addText(Text* text)
		{
			if (text != nullptr)
			{
				if (!text->face)
				{
					throw Util::Exception("Null text face!", KRY_EXCEPTION_DATA);
				}

				Util::Vector2f pos;
				auto gv = glyphs.insert(std::pair<Text*, std::vector<Sprite*> >(text, std::vector<Sprite*>()));

				for (unsigned int i = 0; i < text->text.getLength(); ++i)
				{
					Sprite* sprite = new Sprite;
					FT_Glyph_Metrics metric = text->face->getMetric(text->text[i]);
					Util::Vector2f offset = Util::Vector2f(metric.horiBearingX >> 6, -metric.horiBearingY >> 6) * text->scale;

					sprite->texture = &text->face->getGlyph(text->text[i]);
					sprite->position = pos + offset;
					//sprite->scale = text->scale;
					sprite->rgba = text->rgba;
					sprite->dimensions = Util::Vector2f(sprite->texture->getDimensions()) * text->scale;

					pos[0] += (metric.horiAdvance >> 6) * text->scale[0];
					text->dimensions[0] = pos[0];
					text->dimensions[1] = std::max(text->dimensions[1], sprite->dimensions[1]);
					gv.first->second.push_back(sprite);
				}

				typeBatch.push_back(GraphicObjectType::TEXT);
				graphicBatch.push_back(static_cast<void*>(text));
			}
		}

		void Canvas::addTexts(const std::vector<Text*>& texts)
		{
			for (auto& text : texts)
			{
				addText(text);
			}
		}

		void Canvas::addQuad(Quad* quad)
		{
			if (quad != nullptr)
			{
				typeBatch.push_back(GraphicObjectType::QUAD);
				graphicBatch.push_back(static_cast<void*>(quad));
			}			
		}

		void Canvas::addQuads(const std::vector<Quad*>& quads)
		{
			if (!quads.empty())
			{
				typeBatch.insert(typeBatch.end(), quads.size(), GraphicObjectType::QUAD);
				graphicBatch.insert(graphicBatch.end(), quads.begin(), quads.end());
			}
		}

		void Canvas::addLine(Line* line)
		{
			if (line != nullptr)
			{
				typeBatch.push_back(GraphicObjectType::LINE);
				graphicBatch.push_back(static_cast<void*>(line));
			}			
		}

		void Canvas::addLines(const std::vector<Line*>& lines)
		{
			if (!lines.empty())
			{
				typeBatch.insert(typeBatch.end(), lines.size(), GraphicObjectType::LINE);
				graphicBatch.insert(graphicBatch.end(), lines.begin(), lines.end());
			}
		}

		void Canvas::addTriangle(Triangle* triangle)
		{
			if (triangle != nullptr)
			{
				typeBatch.push_back(GraphicObjectType::TRIANGLE);
				graphicBatch.push_back(static_cast<void*>(triangle));
			}			
		}

		void Canvas::addTriangles(const std::vector<Triangle*>& triangles)
		{
			if (!triangles.empty())
			{
				typeBatch.insert(typeBatch.end(), triangles.size(), GraphicObjectType::TRIANGLE);
				graphicBatch.insert(graphicBatch.end(), triangles.begin(), triangles.end());
			}
		}

		void Canvas::addEllipse(Ellipse* ellipse)
		{
			if (ellipse != nullptr)
			{
				typeBatch.push_back(GraphicObjectType::ELLIPSE);
				graphicBatch.push_back(static_cast<void*>(ellipse));
			}
		}

		void Canvas::addEllipses(const std::vector<Ellipse*>& ellipses)
		{
			if (!ellipses.empty())
			{
				typeBatch.insert(typeBatch.end(), ellipses.size(), GraphicObjectType::ELLIPSE);
				graphicBatch.insert(graphicBatch.end(), ellipses.begin(), ellipses.end());
			}
		}

		void Canvas::setResolution(const Util::Vector2i& Resolution)
		{
			if (Resolution == resolution)
				return;

			resolution = Resolution;
			deleteBuffers();

			// Initialise color buffer
			glGenTextures(1, &colorID);
			glBindTexture(GL_TEXTURE_2D, colorID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, resolution[0], resolution[1], 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
			glBindTexture(GL_TEXTURE_2D, 0);

			// Initialise frame buffer
			glGenFramebuffersEXT(1, &fboID);
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboID);
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, colorID, 0);

			if (glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT) 
			{
				throw Util::Exception("FrameBufferObject not ready!", KRY_EXCEPTION_DATA);
			}

			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
			glFinish();
		}

		Util::Vector2i Canvas::getResolution() const 
		{
			return resolution;
		}

		Util::Vector2f Canvas::getCoord(const Util::Vector2f& Coord) const
		{
			return (Coord * resolution / dimensions - pan - position) / scale;
		}
	}
}