#include "Graphics\Font.h"
#include "Graphics\Texture.h"
#include "Utilities\String.h"
#include "Utilities\Exception.h"
#include <GL\glew.h>
#include <GL\gl.h>
#include <GL\glu.h>

namespace kry
{
	namespace Graphics
	{
		FT_Library Font::library = nullptr;
		Process::Mutex Font::libraryMutex;

		Font::Font()
		{
			libraryMutex.lock();
			init();
			libraryMutex.unlock();
		}

		Font::~Font()
		{
			libraryMutex.lock();
			if (Font::library != nullptr)
			{
				FT_Done_FreeType(Font::library);
				Font::library = nullptr;
			}
			libraryMutex.unlock();
		}

		void Font::init()
		{
			if (Font::library == nullptr)
			{
				FT_Error status = FT_Init_FreeType(&Font::library);

				if (status != 0)
				{
					throw Util::Exception("Failed to initalise TrueType2! [error=" + Util::String(FT_ErrorString[status].err_msg, strlen(FT_ErrorString[status].err_msg)) + ']', KRY_EXCEPTION_DATA);
				}
			}
		}

		void Font::process(const bool MipMap)
		{
			FT_GlyphSlot slot = face->glyph;

			for (unsigned i = 0; i < 255; ++i)
			{
				FT_UInt index = FT_Get_Char_Index(face, i);

				if (FT_Load_Glyph(face, index, FT_LOAD_DEFAULT) != 0)
					continue;

				FT_Render_Glyph(slot, FT_RENDER_MODE_NORMAL);

				int width = slot->bitmap.width;
				int height = slot->bitmap.rows;
				Texture glyph;
				glyph.dimensions = Util::Vector2i(width, height);
				glyph.isMipMapped = MipMap;

				unsigned int ID;
				glGenTextures(1, &ID);
				glBindTexture(GL_TEXTURE_2D, ID);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
				glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

				if (MipMap)
					gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA, width, height, GL_ALPHA, GL_UNSIGNED_BYTE, slot->bitmap.buffer);
				else
					glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, width, height, 0, GL_ALPHA, GL_UNSIGNED_BYTE, slot->bitmap.buffer);

				glyph.textureID = ID;
				textures.insert(std::pair<char, Texture>(i, glyph));
				metrics.insert(std::pair<char, FT_Glyph_Metrics>(i, slot->metrics));
			}

			glBindTexture(GL_TEXTURE_2D, 0);
			glFinish();
		}

		void Font::loadFontFile(const Util::String& Filepath, const unsigned int FontSize, const bool MipMap)
		{
			libraryMutex.lock();
			init();
			mutex.lock();
			FT_Error status = FT_New_Face(Font::library, Filepath.getData(), 0, &face);
				
			if (status == FT_Err_Unknown_File_Format)
			{
				throw Util::Exception("TrueType2 error, failed to open or read the font file! [file=" + Filepath + ']', KRY_EXCEPTION_DATA);
			}
			else if (status != 0)
			{
				throw Util::Exception("Failed to initalise TrueType2 face! [error=" + Util::String(FT_ErrorString[status].err_msg, strlen(FT_ErrorString[status].err_msg)) + ']', KRY_EXCEPTION_DATA);
			}

			if (FT_Set_Pixel_Sizes(face, 0, FontSize))
			{
				throw Util::Exception("Failed setting face size! [size=" + Util::toString(FontSize) + ']', KRY_EXCEPTION_DATA);
			}

			process(MipMap);
			mutex.unlock();
			libraryMutex.unlock();
		}

		void Font::loadFontData(const std::vector<unsigned char>& TrueType, const unsigned int FontSize, const bool MipMap)
		{
			libraryMutex.lock();
			init();
			mutex.lock();
			FT_Error status = FT_New_Memory_Face(Font::library, &TrueType[0], TrueType.size(), 0, &face);

			if (status != 0)
			{
				throw Util::Exception("Failed to initalise TrueType2 memory face! [error=" + Util::String(FT_ErrorString[status].err_msg, strlen(FT_ErrorString[status].err_msg)) + ']', KRY_EXCEPTION_DATA);
			}

			if (FT_Set_Pixel_Sizes(face, 0, FontSize)) 
			{
				throw Util::Exception("Failed setting face size! [size=" + Util::toString(FontSize) + ']', KRY_EXCEPTION_DATA);
			}

			process(MipMap);
			mutex.unlock();
			libraryMutex.unlock();
		}

		Texture& Font::getGlyph(const char Glyph)
		{
			mutex.lock();
			auto find = textures.find(Glyph);

			if (find == textures.end())
			{
				throw Util::Exception("Could not find glyph! [Glyph=" + Util::String(Glyph) + ']', KRY_EXCEPTION_DATA);
			}

			mutex.unlock();
			return find->second;
		}

		FT_Glyph_Metrics Font::getMetric(const char Glyph)
		{
			mutex.lock();
			auto find = metrics.find(Glyph);

			if (find == metrics.end())
			{
				throw Util::Exception("Could not find glyph metric! [Glyph=" + Util::String(Glyph) + ']', KRY_EXCEPTION_DATA);
			}

			mutex.unlock();
			return find->second;
		}

	}
}