#include "Graphics\Renderer.h"
#include "Graphics\Canvas.h"
#include "Graphics\Font.h"
#include "Graphics\Primitives.h"
#include "Graphics\Texture.h"
#include "Utilities\String.h"
#include "Utilities\Exception.h"
#include "Utilities\ErrorMessage.h"
#include "Utilities\Vector.h"
#include "Utilities\Math.h"
#include "System\Window.h"
#include <Windows.h>
#undef min
#undef max
#include <GL\glew.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <vector>
#include <map>

namespace kry
{
	namespace Graphics
	{
		Renderer::Renderer()
		{
			glClearColor(0, 0, 0, 1);
		}

		Renderer::~Renderer()
		{
			clear();
		}

		void Renderer::clear()
		{
			batch.clear();
		}

		void Renderer::clearBuffer(const Util::Vector4f& RGBA)
		{
			glClearColor(RGBA[0], RGBA[1], RGBA[2], RGBA[3]);
			glClear(GL_COLOR_BUFFER_BIT);
		}

		void Renderer::render(System::DeviceContext DeviceContext, System::GLContext GLContext, const Util::Vector2i& RenderDimensions)
		{
			auto currentContext = wglGetCurrentContext();
			auto currentDC = wglGetCurrentDC();

			if (wglMakeCurrent(DeviceContext, GLContext) == FALSE)
			{
				throw Util::Exception("Failed to switch render context! [error=" + Util::parseWinError(GetLastError()) + ']', KRY_EXCEPTION_DATA);
			}		

			glEnable(GL_BLEND);
			glBlendFuncSeparateEXT(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA, GL_DST_ALPHA);
			glBlendEquationSeparateEXT(GL_FUNC_ADD, GL_FUNC_ADD);
			Canvas* previousCanvas = nullptr;

			for (auto& canvas : batch)
			{
				auto canvasTypeBatch = &canvas->typeBatch;
				auto canvasGraphicBatch = &canvas->graphicBatch;

				if (!canvas->fboID || !canvas->colorID || canvas->resolution <= Util::Vector2i(0))
				{
					throw Util::Exception("Canvas resolution needs to be set!", KRY_EXCEPTION_DATA);
				}

				if (canvas->rgba[3] <= 0.0f || canvas->scale == Util::Vector2f(0))
					continue;

				if (canvasTypeBatch->size() != canvasGraphicBatch->size())
				{
					throw Util::Exception("Canvas batch alignment error!", KRY_EXCEPTION_DATA);
				}

				// Render canvas to FBO
				glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, canvas->fboID);
				glClear(GL_COLOR_BUFFER_BIT);

				// First add previous canvas in background (otherwise clear color of this canvas will be used to blend void, so blend with previous canvas)
				if (previousCanvas != nullptr)
				{
					glMatrixMode(GL_PROJECTION);
					glLoadIdentity();
					glViewport(0, 0, canvas->resolution[0], canvas->resolution[1]);
					glOrtho(0, canvas->resolution[0], canvas->resolution[1], 0, -1, 1);
					glMatrixMode(GL_MODELVIEW);
					glLoadIdentity();
					glEnable(GL_TEXTURE_2D);
					Util::Vector2f offset = previousCanvas->dimensions * Util::Vector2f(0.5);
					glTranslatef(previousCanvas->position[0] + offset[0], previousCanvas->position[1] + offset[1], 0);

					if (previousCanvas->rotation != 0.0f)
						glRotatef(previousCanvas->rotation, 0, 0, 1);

					/** #TODO(fix) Blending RGBA effects previous canvases too */
					glColor4fv(&previousCanvas->rgba[0]);
					glBindTexture(GL_TEXTURE_2D, previousCanvas->colorID);
					GLint filter = previousCanvas->linearFilter ? GL_LINEAR : GL_NEAREST;
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);

					glBegin(GL_QUADS);
					glTexCoord2f(0, 1);
					glVertex2f(-offset[0], -offset[1]);
					glTexCoord2f(1, 1);
					glVertex2f(offset[0], -offset[1]);
					glTexCoord2f(1, 0);
					glVertex2f(offset[0], offset[1]);
					glTexCoord2f(0, 0);
					glVertex2f(-offset[0], offset[1]);
					glEnd();

					glDisable(GL_TEXTURE_2D);
				}

				previousCanvas = canvas;

				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glViewport(0, 0, canvas->resolution[0], canvas->resolution[1]);
				glOrtho(0, canvas->resolution[0], canvas->resolution[1], 0, -1, 1);
				glTranslatef(canvas->pan[0], canvas->pan[1], 1);
				glScalef(canvas->scale[0], canvas->scale[1], 1);
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();

				// Render canvas batch
				auto& type = canvasTypeBatch->begin();
				auto& graphic = canvasGraphicBatch->begin();
				for (; type != canvasTypeBatch->end(); ++type, ++graphic)
				{
					glLoadIdentity();

					switch (*type)
					{
					case GraphicObjectType::SPRITE: renderSprite(static_cast<Sprite*>(*graphic)); break;
					case GraphicObjectType::TEXT: renderText(static_cast<Text*>(*graphic), canvas); break;
					case GraphicObjectType::QUAD: renderQuad(static_cast<Quad*>(*graphic)); break;
					case GraphicObjectType::LINE: renderLine(static_cast<Line*>(*graphic)); break;
					case GraphicObjectType::TRIANGLE: renderTriangle(static_cast<Triangle*>(*graphic)); break;
					case GraphicObjectType::ELLIPSE: renderEllipse(static_cast<Ellipse*>(*graphic)); break;
					}
				}
			}

			// Render final output
			if (previousCanvas != nullptr)
			{
				glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glViewport(0, 0, RenderDimensions[0], RenderDimensions[1]);
				glOrtho(0, RenderDimensions[0], RenderDimensions[1], 0, -1, 1);
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();
				glEnable(GL_TEXTURE_2D);
				Util::Vector2f offset = previousCanvas->dimensions * Util::Vector2f(0.5);
				glTranslatef(previousCanvas->position[0] + offset[0], previousCanvas->position[1] + offset[1], 0);

				if (previousCanvas->rotation != 0.0f)
					glRotatef(previousCanvas->rotation, 0, 0, 1);

				glColor4fv(&previousCanvas->rgba[0]);
				glBindTexture(GL_TEXTURE_2D, previousCanvas->colorID);
				GLint filter = previousCanvas->linearFilter ? GL_LINEAR : GL_NEAREST;
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);

				glBegin(GL_QUADS);
				glTexCoord2f(0, 1);
				glVertex2f(-offset[0], -offset[1]);
				glTexCoord2f(1, 1);
				glVertex2f(offset[0], -offset[1]);
				glTexCoord2f(1, 0);
				glVertex2f(offset[0], offset[1]);
				glTexCoord2f(0, 0);
				glVertex2f(-offset[0], offset[1]);
				glEnd();

				glDisable(GL_TEXTURE_2D);
			}

			glFlush();
			SwapBuffers(DeviceContext);

			if (wglMakeCurrent(currentDC, currentContext) == FALSE)
			{
				throw Util::Exception("Failed to switch render context back! [error=" + Util::parseWinError(GetLastError()) + ']', KRY_EXCEPTION_DATA);
			}
		}

		void Renderer::addCanvas(Canvas* canvas)
		{
			batch.push_back(canvas);
		}

		void Renderer::renderSprite(Sprite* sprite)
		{
			if (sprite->rgba[3] > 0.0f)
			{
				if (sprite->texture != nullptr)
				{
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, sprite->texture->textureID);

					if (sprite->texture->isMipMapped)
					{
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, sprite->linearFilter ? GL_LINEAR : GL_NEAREST);
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, sprite->linearFilter ? GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST);
					}
					else
					{
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, sprite->linearFilter ? GL_LINEAR : GL_NEAREST);
						glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, sprite->linearFilter ? GL_LINEAR : GL_NEAREST);
					}
				}
					
				glColor4fv(&sprite->rgba[0]);
				Util::Vector2f offset = sprite->dimensions * Util::Vector2f(0.5);
				glTranslatef(sprite->position[0] + offset[0], sprite->position[1] + offset[1], 0);
				glRotatef(sprite->rotation, 0, 0, 1);
				glScalef(sprite->scale[0], sprite->scale[1], 1);

				glBegin(GL_QUADS);
				// Top-left (clockwise)
				glTexCoord2f(sprite->uv[0][0], sprite->uv[0][1]);
				glVertex2f(-offset[0], -offset[1]);
				glTexCoord2f(sprite->uv[3][0], sprite->uv[3][1]);
				glVertex2f(offset[0], -offset[1]);
				glTexCoord2f(sprite->uv[2][0], sprite->uv[2][1]);
				glVertex2f(offset[0], offset[1]);
				glTexCoord2f(sprite->uv[1][0], sprite->uv[1][1]);
				glVertex2f(-offset[0], offset[1]);
				glEnd();

				glDisable(GL_TEXTURE_2D);
			}
		}

		void Renderer::renderText(Text* text, Canvas* canvas)
		{
			auto textBatch = canvas->glyphs.find(text);

			if (textBatch != canvas->glyphs.end())
			{
				for (auto& i : textBatch->second)
				{
					glLoadIdentity();
					glRotatef(text->rotation, 0, 0, 1);
					i->position += text->position;
					renderSprite(i);
					i->position -= text->position;
				}					
			}
		}

		void Renderer::renderQuad(Quad* quad)
		{
			float lineThickness;
			glGetFloatv(GL_LINE_WIDTH, &lineThickness);
			glLineWidth(quad->thickness);

			if (quad->rgba[3] > 0.0f)
			{
				glColor4fv(&quad->rgba[0]);
				Util::Vector2f offset = quad->dimensions * Util::Vector2f(0.5);				
				glTranslatef(quad->position[0] + offset[0], quad->position[1] + offset[1], 0);
				glRotatef(quad->rotation, 0, 0, 1);
				glScalef(quad->scale[0], quad->scale[1], 1);

				if (quad->thickness <= 0.0f) // Fill
					glBegin(GL_QUADS);
				else
					glBegin(GL_LINE_LOOP); // Border

				glVertex2f(-offset[0], -offset[1]);
				glVertex2f(offset[0], -offset[1]);
				glVertex2f(offset[0], offset[1]);
				glVertex2f(-offset[0], offset[1]);
				glEnd();
			}

			glLineWidth(lineThickness);
		}

		void Renderer::renderLine(Line* line)
		{
			float lineThickness;
			glGetFloatv(GL_LINE_WIDTH, &lineThickness);
			glLineWidth(line->thickness);

			if (line->rgba[3] > 0.0f)
			{
				glColor4fv(&line->rgba[0]);
				glScalef(line->scale[0], line->scale[1], 1);

				glBegin(GL_LINES);
				glVertex2f(line->points[0][0], line->points[0][1]);
				glVertex2f(line->points[1][0], line->points[1][1]);
				glEnd();
			}

			glLineWidth(lineThickness);
		}

		void Renderer::renderTriangle(Triangle* triangle)
		{
			float lineThickness;
			glGetFloatv(GL_LINE_WIDTH, &lineThickness);
			glLineWidth(triangle->thickness);

			if (triangle->rgba[3] > 0.0f)
			{
				glColor4fv(&triangle->rgba[0]);
				Util::Vector2f center = (triangle->points[0] + triangle->points[1] + triangle->points[2]) / Util::Vector2f(3);				
				glTranslatef(center[0], center[1], 0);
				glRotatef(triangle->rotation, 0, 0, 1);
				glScalef(triangle->scale[0], triangle->scale[1], 1);

				if (triangle->thickness <= 0.0f) // Fill
					glBegin(GL_TRIANGLES);
				else
					glBegin(GL_LINE_LOOP); // Border

				glVertex2f(triangle->points[0][0] - center[0], triangle->points[0][1] - center[1]);
				glVertex2f(triangle->points[1][0] - center[0], triangle->points[1][1] - center[1]);
				glVertex2f(triangle->points[2][0] - center[0], triangle->points[2][1] - center[1]);
				glEnd();
			}

			glLineWidth(lineThickness);
		}

		void Renderer::renderEllipse(Ellipse* ellipse)
		{
			float lineThickness;
			glGetFloatv(GL_LINE_WIDTH, &lineThickness);
			glLineWidth(ellipse->thickness);

			if (ellipse->rgba[3] > 0.0f)
			{
				glColor4fv(&ellipse->rgba[0]);
				Util::Vector2f offset = ellipse->radius * Util::Vector2f(0.5);				
				glTranslatef(ellipse->position[0] + offset[0], ellipse->position[1] + offset[1], 0);
				glRotatef(ellipse->rotation, 0, 0, 1);
				glScalef(ellipse->scale[0], ellipse->scale[1], 1);					

				if (ellipse->thickness <= 0.0f)
				{ 
					// Fill
					glBegin(GL_TRIANGLE_FAN);
					glVertex2f(0, 0);

					for (int i = ellipse->segments; i >= 0; --i)
					{
						float radians = ((float)i / ellipse->segments) * (KRY_PI * 2);
						float segment_x = ellipse->radius[0] * cos(radians);
						float segment_y = ellipse->radius[1] * sin(radians);
						glVertex2f(segment_x, segment_y);
					}
					glEnd();
				}
				else 
				{ 
					// Border
					glBegin(GL_LINE_LOOP);
					for (unsigned i = 0; i < ellipse->segments; ++i) 
					{
						float radians = ((float)i / ellipse->segments) * (KRY_PI * 2);
						float segment_x = ellipse->radius[0] * cos(radians);
						float segment_y = ellipse->radius[1] * sin(radians);
						glVertex2f(segment_x, segment_y);
					}
					glEnd();
				}
			}

			glLineWidth(lineThickness);
		}
	}
}