#include "Media\AudioFactory.h"
#include "Utilities\Exception.h"
#include "Utilities\String.h"
#include "Utilities\Math.h"
#include "Audio\Audio.h"
#include "..\libs\include\mpg123\mpg123.h"
#include <string>
#include <algorithm>
#include <fstream>
#include <Windows.h>

#define KRY_AUDIO_BUFFER_SIZE 1024
#define KRY_AUDIO_PCM_SIZE 4096
#define KRY_AUDIO_BUFFERS 4

namespace kry
{
	namespace Media
	{
		struct Mpg123Manager
		{
			Mpg123Manager()
			{
				active = false;
			}

			~Mpg123Manager()
			{
				if (active)
				{
					mpg123_exit();
				}
			}

			void init()
			{
				if (!active)
				{
					mpg123_init();
					active = true;
				}
			}

		private:
			bool active;
		};

		static Mpg123Manager mpg123Manager = Mpg123Manager();

		struct mp3Instance
		{
			mpg123_handle* handle = nullptr;
			int bits = 0;
			int channels = 0;
			int rate = 0;
		};

		std::vector<unsigned char> mp3BufferPCM(mp3Instance& mp3, const unsigned char* Buffer, const size_t Length, Audio::Format& Format, unsigned int& Frequency)
		{
			size_t size = 0;
			std::vector<unsigned char> pcmVector;
			unsigned char pcm[KRY_AUDIO_PCM_SIZE];

			int status = mpg123_decode(mp3.handle, Buffer, Length, 0, 0, &size);

			if (status == MPG123_NEW_FORMAT)
			{
				int encoding;
				mpg123_getformat(mp3.handle, reinterpret_cast<long*>(&mp3.rate), &mp3.channels, &encoding);

				switch (encoding)
				{
				case MPG123_ENC_SIGNED_8: mp3.bits = 8; break;
				case MPG123_ENC_UNSIGNED_8: mp3.bits = 8; break;
				case MPG123_ENC_SIGNED_16: mp3.bits = 16; break;
				case MPG123_ENC_UNSIGNED_16: mp3.bits = 16; break;
				case MPG123_ENC_SIGNED_24: mp3.bits = 24; break;
				case MPG123_ENC_UNSIGNED_24: mp3.bits = 24; break;
				case MPG123_ENC_SIGNED_32: mp3.bits = 32; break;
				case MPG123_ENC_UNSIGNED_32: mp3.bits = 32; break;
				case MPG123_ENC_FLOAT_32: mp3.bits = 32; break;
				}

				Frequency = mp3.rate;

				/** #TODO(incomplete) add support for all types of audio */
				if (mp3.bits == 8)
				{
					switch (mp3.channels)
					{
					case 1: Format = Audio::Format::MONO_8; break;
					case 2: Format = Audio::Format::STEREO_8; break;
					}
				}
				else if (mp3.bits == 16)
				{
					switch (mp3.channels)
					{
					case 1: Format = Audio::Format::MONO_16; break;
					case 2: Format = Audio::Format::STEREO_16; break;
					case 4: Format = Audio::Format::QUAD_16; break;
					case 5: Format = Audio::Format::CHANNEL_51_16; break;
					case 6: Format = Audio::Format::CHANNEL_61_16; break;
					case 7: Format = Audio::Format::CHANNEL_71_16; break;
					}
				}
				else if (mp3.bits == 24)
				{

				}
				else if (mp3.bits == 32)
				{

				}
			}

			while (status != MPG123_ERR && status != MPG123_NEED_MORE)
			{
				status = mpg123_decode(mp3.handle, 0, 0, pcm, sizeof(pcm), &size);

				if (size > 0)
				{
					pcmVector.insert(pcmVector.end(), pcm, pcm + size);
				}
			}

			if (status == MPG123_ERR)
			{
				throw Util::Exception("MP3 decoder error! [error=" + Util::String(mpg123_plain_strerror(status), strlen(mpg123_plain_strerror(status))) + ']', KRY_EXCEPTION_DATA);
			}

			return pcmVector;
		}

		std::vector<unsigned char> mp3FileToPCM(const Util::String& Filepath, Audio::Format& Format, unsigned int& Frequency)
		{
			mpg123Manager.init();
			std::vector<unsigned char> pcmVector;
			unsigned char buffer[KRY_AUDIO_BUFFER_SIZE];
			size_t pos = 0;
			size_t length = 0;
			mp3Instance mp3;

			FILE *fp = 0;
			fopen_s(&fp, Filepath.getData(), "rb");

			if (!fp)
			{
				throw Util::Exception("MP3 decoder error, could not open file! [file=" + Filepath + ']', KRY_EXCEPTION_DATA);
			}
			else
			{
				fseek(fp, 0, SEEK_END);
				size_t fileSize = ftell(fp);
				rewind(fp);

				int status;
				mp3.handle = mpg123_new(0, &status);

				if (mp3.handle == nullptr || mp3.handle == 0)
				{
					throw Util::Exception("MP3 decoder error! [error=" + Util::String(mpg123_plain_strerror(status), strlen(mpg123_plain_strerror(status))) + ']', KRY_EXCEPTION_DATA);
				}
				else
				{
					mpg123_open_feed(mp3.handle);

					// Decode MP3
					while (pos < fileSize)
					{
						length = fread(reinterpret_cast<void*>(&buffer[0]), sizeof(unsigned char), sizeof(buffer), fp);
						std::vector<unsigned char> pcmBuffer = mp3BufferPCM(mp3, buffer, length, Format, Frequency);
						pcmVector.insert(pcmVector.end(), pcmBuffer.begin(), pcmBuffer.end());
						pos += length;
					}
				}

				mpg123_delete(mp3.handle);
			}

			fclose(fp);

			return pcmVector;
		}

		std::vector<unsigned char> mp3DataToPCM(const std::vector<unsigned char>& Data, Audio::Format& Format, unsigned int& Frequency)
		{
			mpg123Manager.init();
			std::vector<unsigned char> pcmVector;
			unsigned char* buffer = nullptr;
			size_t pos = 0;
			size_t length = 0;
			mp3Instance mp3;

			int status;
			mp3.handle = mpg123_new(0, &status);

			if (mp3.handle == nullptr || mp3.handle == 0)
			{
				throw Util::Exception("MP3 decoder error! [error=" + Util::String(mpg123_plain_strerror(status), strlen(mpg123_plain_strerror(status))) + ']', KRY_EXCEPTION_DATA);
			}
			else
			{
				mpg123_open_feed(mp3.handle);

				// Decode MP3
				while (pos < Data.size())
				{
					buffer = const_cast<unsigned char*>(&Data[pos]);
					length = Util::clampi(static_cast<int>(Data.size()) - static_cast<int>(pos), 0, KRY_AUDIO_BUFFER_SIZE);
					std::vector<unsigned char> pcmBuffer = mp3BufferPCM(mp3, buffer, length, Format, Frequency);
					pcmVector.insert(pcmVector.end(), pcmBuffer.begin(), pcmBuffer.end());
					pos += length;
				}

				mpg123_delete(mp3.handle);
			}

			return pcmVector;
		}

		void mp3FileStream(const Util::String& Filepath, Audio::Source& Source, const bool& Terminate)
		{
			mpg123Manager.init();
			unsigned char buffer[KRY_AUDIO_BUFFER_SIZE];
			size_t pos = 0;
			size_t length = 0;
			Audio::Format format;
			unsigned int frequency = 0;
			mp3Instance mp3;
			Audio::Buffer queue[KRY_AUDIO_BUFFERS];
			int queueIndex = 0;
			int queueCount = 1;

			FILE *fp = 0;
			fopen_s(&fp, Filepath.getData(), "rb");

			if (!fp)
			{
				throw Util::Exception("MP3 decoder error, could not open file! [file=" + Filepath + ']', KRY_EXCEPTION_DATA);
			}

			fseek(fp, 0, SEEK_END);
			size_t fileSize = ftell(fp);
			rewind(fp);

			int status;
			mp3.handle = mpg123_new(0, &status);

			if (mp3.handle == nullptr || mp3.handle == 0)
			{
				throw Util::Exception("MP3 decoder error! [error=" + Util::String(mpg123_plain_strerror(status), strlen(mpg123_plain_strerror(status))) + ']', KRY_EXCEPTION_DATA);
			}

			mpg123_open_feed(mp3.handle);
			Source.bind();

			// Decode MP3
			while (pos < fileSize)
			{
				length = fread(reinterpret_cast<void*>(&buffer[0]), sizeof(unsigned char), sizeof(buffer), fp);
				std::vector<unsigned char> pcmBuffer = mp3BufferPCM(mp3, buffer, length, format, frequency);

				if (pcmBuffer.size() > 0)
				{
					if (Terminate)
						break;

					// Wait for free buffer
					while (Source.getStatus() != Audio::Status::STOPPED && Source.getCurrBuffer() == &queue[queueIndex])
						Sleep(10);

					if (queueCount > KRY_AUDIO_BUFFERS)
					{
						Source.unqueue(queue[queueIndex]);

						if (Source.getStatus() == Audio::Status::STOPPED)
							break;
					}
					else
					{
						++queueCount;

						if (queueCount == KRY_AUDIO_BUFFERS)
							Source.play();
					}

					// Queue buffer
					queue[queueIndex].resetData(pcmBuffer, format, frequency);
					Source.queue(queue[queueIndex]);
					queueIndex = (queueIndex + 1) % KRY_AUDIO_BUFFERS;
				}

				pos += length;
			}

			// Wait for playback to finish
			while (Source.getStatus() == Audio::Status::PLAYING)
				Sleep(10);

			Source.clearQueue();
			Source.unbind();
			mpg123_delete(mp3.handle);

			fclose(fp);
		}

		void mp3DataStream(const std::vector<unsigned char>& Data, Audio::Source& Source, const bool& Terminate)
		{
			mpg123Manager.init();
			unsigned char* buffer = nullptr;
			size_t pos = 0;
			size_t length = 0;
			Audio::Format format;
			unsigned int frequency = 0;
			mp3Instance mp3;
			Audio::Buffer queue[KRY_AUDIO_BUFFERS];
			int queueIndex = 0;
			int queueCount = 1;

			int status;
			mp3.handle = mpg123_new(0, &status);

			if (mp3.handle == nullptr || mp3.handle == 0)
			{
				throw Util::Exception("MP3 decoder error! [error=" + Util::String(mpg123_plain_strerror(status), strlen(mpg123_plain_strerror(status))) + ']', KRY_EXCEPTION_DATA);
			}

			mpg123_open_feed(mp3.handle);
			Source.bind();

			// Decode MP3
			while (pos < Data.size())
			{
				buffer = const_cast<unsigned char*>(&Data[pos]);
				length = Util::clampi(static_cast<int>(Data.size()) - static_cast<int>(pos), 0, KRY_AUDIO_BUFFER_SIZE);
				std::vector<unsigned char> pcmBuffer = mp3BufferPCM(mp3, buffer, length, format, frequency);

				if (pcmBuffer.size() > 0)
				{
					if (Terminate)
						break;

					// Wait for free buffer
					while (Source.getStatus() != Audio::Status::STOPPED && Source.getCurrBuffer() == &queue[queueIndex])
						Sleep(10);

					if (queueCount > KRY_AUDIO_BUFFERS)
					{
						Source.unqueue(queue[queueIndex]);

						if (Source.getStatus() == Audio::Status::STOPPED)
							break;
					}
					else
					{
						++queueCount;

						if (queueCount == KRY_AUDIO_BUFFERS)
							Source.play();
					}

					// Queue buffer
					queue[queueIndex].resetData(pcmBuffer, format, frequency);
					Source.queue(queue[queueIndex]);
					queueIndex = (queueIndex + 1) % KRY_AUDIO_BUFFERS;
				}

				pos += length;
			}

			// Wait for playback to finish
			while (Source.getStatus() == Audio::Status::PLAYING)
				Sleep(10);

			Source.clearQueue();
			Source.unbind();
			mpg123_delete(mp3.handle);
		}
	}
}