#include "Media\ConfigFactory.h"
#include "Utilities\String.h"
#include "Utilities\Exception.h"
#include <fstream>
#include <string>

namespace kry
{
	namespace Media
	{
		Config::Config()
		{

		}

		Config::Config(const Util::String& Filepath)
		{
			fileToConfig(Filepath);
		}

		Config::Config(const std::vector<Util::String>& Lines)
		{
			linesToConfig(Lines);
		}

		Config::Config(const std::vector<unsigned char>& Data)
		{
			dataToConfig(Data);
		}

		Config::~Config()
		{
			clear();
		}

		void Config::fileToConfig(const Util::String& Filepath)
		{
			std::ifstream fin;
			fin.open(Filepath.getData(), std::fstream::in);

			if (!fin.good())
			{
				throw Util::Exception("Config error, could not open file! [file=" + Filepath + ']', KRY_EXCEPTION_DATA);
			}
			else
			{
				std::string line;
				std::vector<Util::String> lines;

				while (fin.good())
				{
					std::getline(fin, line);
					lines.push_back(Util::String(&line[0], line.size()));
				}

				linesToConfig(lines);
			}

			fin.close();
		}

		void Config::dataToConfig(const std::vector<unsigned char>& Data)
		{
			std::vector<Util::String> lines;
			unsigned long left = 0;

			for (unsigned long i = 0; i < Data.size(); ++i)
			{
				if (Data[i] == '\r' || Data[i] == '\n' || Data[i] == '\0')
				{
					if (Data[i] == '\r')
					{
						lines.push_back(Util::String(reinterpret_cast<const char*>(&Data[left]), i - left));
						++i;
						left = i + 1;
					}
					else
					{
						lines.push_back(Util::String(reinterpret_cast<const char*>(&Data[left]), i - left));
						left = i + 1;
					}
				}
			}

			if (left < Data.size() - 1)
			{
				lines.push_back(Util::String(reinterpret_cast<const char*>(&Data[left]), Data.size() - left));
			}

			linesToConfig(lines);
		}
		
		void Config::linesToConfig(const std::vector<Util::String>& Lines)
		{
			ConfigSection* currentSection = &config[Util::String()];

			for (auto& line : Lines)
			{
				int sectionStart = 0;
				bool isSection = false;
				bool isKey = false;

				for (unsigned int i = 0; i < line.getLength(); ++i)
				{
					if (!isSection && !isKey && line[i] == ';')
						break;

					if (!isSection && !isKey && line[i] == '[')
					{
						sectionStart = i + 1;
						isSection = true;
					}

					if (isSection && line[i] == ']')
					{
						currentSection = &config[trim(Util::String(&line[sectionStart], i - sectionStart))];
					}

					if (!isSection && line[i] == '=')
					{
						isKey = true;
						(*currentSection)[trim(Util::String(&line[0], i))] = trim(Util::String(&line[i + 1], line.getLength() - (i + 1)));
					}
				}
			}
		}

		void Config::clear()
		{
			config.clear();
		}

		void Config::removeSection(const Util::String& Section)
		{
			config.erase(Section);
		}

		std::vector<Util::String> Config::getSectionNames() const
		{
			std::vector<Util::String> names;

			for (auto& name : config)
			{
				names.push_back(name.first);
			}

			return names;
		}

		unsigned int Config::getSectionsSize() const
		{
			return config.size();
		}

		bool Config::sectionExists(const Util::String& Section) const
		{
			return (config.find(Section) != config.end());
		}

		Config::ConfigSection& Config::operator[](const Util::String& Section)
		{
			auto i = config.find(Section);

			if (i == config.end())
			{
				return config.insert(std::pair<Util::String, ConfigSection>(Section, ConfigSection())).first->second;
			}

			return i->second;
		}

		Util::String Config::trim(const Util::String& Dirty)
		{
			int left = 0;
			int right = Dirty.getLength();

			if (right > 0)
			{
				for (unsigned int i = 0; i < Dirty.getLength(); ++i)
				{
					if (Dirty[i] != ' ' && Dirty[i] != '\t')
					{
						left = i;
						break;
					}
				}

				for (int i = int(Dirty.getLength()) - 1; i >= 0; --i)
				{
					if (Dirty[i] != ' ' && Dirty[i] != '\t')
					{
						right = i + 1;
						break;
					}
				}

				return Util::String(&Dirty[left], right - left);
			}

			return Dirty;
		}

		/* Config::ConfigSection */

		void Config::ConfigSection::clear()
		{
			keysValues.clear();
		}

		void Config::ConfigSection::removeKey(const Util::String& Key)
		{
			keysValues.erase(Key);
		}
		
		std::vector<Util::String> Config::ConfigSection::getKeyNames() const
		{
			std::vector<Util::String> names;

			for (auto& name : keysValues)
			{
				names.push_back(name.first);
			}

			return names;
		}

		unsigned int Config::ConfigSection::getKeysSize() const
		{
			return keysValues.size();
		}

		bool Config::ConfigSection::keyExists(const Util::String& Key) const
		{
			return (keysValues.find(Key) != keysValues.end());
		}

		Util::String& Config::ConfigSection::operator[](const Util::String& Key)
		{
			auto i = keysValues.find(Key);

			if (i == keysValues.end())
			{
				return keysValues.insert(std::pair<Util::String, Util::String>(Key, Util::String())).first->second;
			}

			return i->second;
		}

	}
}