#include "Media\Zip.h"
#include "Utilities\Exception.h"
#include "../miniz/miniz.c"
#include <fstream>

namespace kry
{
	namespace Media
	{
		Zip::Zip()
		{

		}
		
		Zip::Zip(const Util::String& Filepath)
		{
			fileToZip(Filepath);
		}

		Zip::Zip(const std::vector<unsigned char>& Data)
		{
			dataToZip(Data);
		}

		Zip::~Zip()
		{
			clear();
		}


		void Zip::clear()
		{
			zip.clear();
		}

		void Zip::fileToZip(const Util::String& ZipAddress)
		{
			mz_zip_archive archive;
			memset(&archive, 0, sizeof(mz_zip_archive));

			if (!mz_zip_reader_init_file(&archive, ZipAddress.getData(), 0))
			{
				throw Util::Exception("Zip error, could not open file! [address=" + ZipAddress + ']', KRY_EXCEPTION_DATA);
			}

			int files = mz_zip_reader_get_num_files(&archive);

			for (int i = 0; i < files; ++i)
			{
				mz_zip_archive_file_stat fileInfo;

				if (!mz_zip_reader_file_stat(&archive, i, &fileInfo))
				{
					mz_zip_reader_end(&archive);
					throw Util::Exception("Zip error, file stat failed!", KRY_EXCEPTION_DATA);
				}

				Util::String filename(fileInfo.m_filename, strlen(fileInfo.m_filename));
				int directoryPos = filename.lastIndexOf('/') + 1;
				int filePos = filename.getLength() - directoryPos;

				if (filePos > 0)
				{
					auto& file = zip[filename.substring(0, directoryPos)][filename.substring(directoryPos, filePos)];
					unsigned int rawSize = fileInfo.m_uncomp_size;

					void* data = mz_zip_reader_extract_file_to_heap(&archive, filename.getData(), &rawSize, 0);
					if (!data)
					{
						mz_zip_reader_end(&archive);
						throw Util::Exception("Zip error, file extraction failed (returned null)!", KRY_EXCEPTION_DATA);
					}

					file.clear();
					file.insert(file.begin(), static_cast<unsigned char*>(data), static_cast<unsigned char*>(data) + rawSize);
					mz_free(data);
				}
				else
				{
					zip[filename.substring(0, directoryPos)];
				}
			}

			mz_zip_reader_end(&archive);
		}

		void Zip::dataToZip(const std::vector<unsigned char>& Data)
		{
			mz_zip_archive archive;
			memset(&archive, 0, sizeof(mz_zip_archive));

			if (!mz_zip_reader_init_mem(&archive, static_cast<const void*>(&Data[0]), Data.size(), 0))
			{
				throw Util::Exception("Zip error, could not interpret data!", KRY_EXCEPTION_DATA);
			}

			int files = mz_zip_reader_get_num_files(&archive);

			for (int i = 0; i < files; ++i)
			{
				mz_zip_archive_file_stat fileInfo;

				if (!mz_zip_reader_file_stat(&archive, i, &fileInfo))
				{
					mz_zip_reader_end(&archive);
					throw Util::Exception("Zip error, file stat failed!", KRY_EXCEPTION_DATA);
				}

				Util::String filename(fileInfo.m_filename, strlen(fileInfo.m_filename));
				int directoryPos = filename.lastIndexOf('/') + 1;
				int filePos = filename.getLength() - directoryPos;

				if (filePos > 0)
				{
					auto& file = zip[filename.substring(0, directoryPos)][filename.substring(directoryPos, filePos)];
					unsigned int rawSize = 0;
					void* data = mz_zip_reader_extract_to_heap(&archive, fileInfo.m_file_index, &rawSize, 0);

					if (!data)
					{
						mz_zip_reader_end(&archive);
						throw Util::Exception("Zip error, file extraction failed (returned null)!", KRY_EXCEPTION_DATA);
					}

					file.clear();
					file.insert(file.begin(), static_cast<unsigned char*>(data), static_cast<unsigned char*>(data) + rawSize);
					mz_free(data);
				}
				else
				{
					zip[filename.substring(0, directoryPos)];
				}
			}

			mz_zip_reader_end(&archive);
		}

		std::vector<unsigned char> Zip::fileFromFileZip(const Util::String& ZipAddress, const Util::String& FilePath)
		{
			std::vector<unsigned char> file;
			bool found = false;
			int directoryPos = FilePath.lastIndexOf('/') + 1;
			int filePos = FilePath.getLength() - directoryPos;

			if (filePos <= 0)
			{
				throw Util::Exception("Cannot extract directory! [path=" + FilePath + ']', KRY_EXCEPTION_DATA);
			}

			mz_zip_archive archive;
			memset(&archive, 0, sizeof(mz_zip_archive));

			if (!mz_zip_reader_init_file(&archive, ZipAddress.getData(), 0))
			{
				throw Util::Exception("Zip error, could not open file! [address=" + ZipAddress + ']', KRY_EXCEPTION_DATA);
			}

			int files = mz_zip_reader_get_num_files(&archive);

			for (int i = 0; i < files; ++i)
			{
				mz_zip_archive_file_stat fileInfo;

				if (!mz_zip_reader_file_stat(&archive, i, &fileInfo))
				{
					mz_zip_reader_end(&archive);
					throw Util::Exception("Zip error, file stat failed!", KRY_EXCEPTION_DATA);
				}

				Util::String filename(fileInfo.m_filename, strlen(fileInfo.m_filename));

				if (filename == FilePath)
				{
					unsigned int rawSize = fileInfo.m_uncomp_size;
					void* data = mz_zip_reader_extract_file_to_heap(&archive, filename.getData(), &rawSize, 0);
					if (!data)
					{
						mz_zip_reader_end(&archive);
						throw Util::Exception("Zip error, file extraction failed (returned null)!", KRY_EXCEPTION_DATA);
					}
										
					file.insert(file.begin(), static_cast<unsigned char*>(data), static_cast<unsigned char*>(data)+rawSize);
					mz_free(data);
					found = true;
					break;
				}
			}

			mz_zip_reader_end(&archive);

			if (!found)
			{
				throw Util::Exception("Zip error, could not find file in zip! [address=" + ZipAddress + "][path=" + FilePath + ']', KRY_EXCEPTION_DATA);
			}

			return file;
		}

		std::vector<unsigned char> Zip::fileFromDataZip(const std::vector<unsigned char>& Data, const Util::String& FilePath)
		{
			std::vector<unsigned char> file;
			bool found = false;
			int directoryPos = FilePath.lastIndexOf('/') + 1;
			int filePos = FilePath.getLength() - directoryPos;

			if (filePos <= 0)
			{
				throw Util::Exception("Cannot extract directory! [path=" + FilePath + ']', KRY_EXCEPTION_DATA);
			}

			mz_zip_archive archive;
			memset(&archive, 0, sizeof(mz_zip_archive));

			if (!mz_zip_reader_init_mem(&archive, static_cast<const void*>(&Data[0]), Data.size(), 0))
			{
				throw Util::Exception("Zip error, could not interpret data!", KRY_EXCEPTION_DATA);
			}

			int files = mz_zip_reader_get_num_files(&archive);

			for (int i = 0; i < files; ++i)
			{
				mz_zip_archive_file_stat fileInfo;

				if (!mz_zip_reader_file_stat(&archive, i, &fileInfo))
				{
					mz_zip_reader_end(&archive);
					throw Util::Exception("Zip error, file stat failed!", KRY_EXCEPTION_DATA);
				}

				Util::String filename(fileInfo.m_filename, strlen(fileInfo.m_filename));

				if (filename == FilePath)
				{
					unsigned int rawSize = 0;
					void* data = mz_zip_reader_extract_to_heap(&archive, fileInfo.m_file_index, &rawSize, 0);

					if (!data)
					{
						mz_zip_reader_end(&archive);
						throw Util::Exception("Zip error, file extraction failed (returned null)!", KRY_EXCEPTION_DATA);
					}

					file.insert(file.begin(), static_cast<unsigned char*>(data), static_cast<unsigned char*>(data)+rawSize);
					mz_free(data);
					found = true;
					break;
				}
			}

			mz_zip_reader_end(&archive);

			if (!found)
			{
				throw Util::Exception("Zip error, could not find file in data zip! [path=" + FilePath + ']', KRY_EXCEPTION_DATA);
			}

			return file;
		}

		std::vector<Util::String> Zip::getDirectoryNames() const
		{
			std::vector<Util::String> names;

			for (auto& name : zip)
			{
				names.push_back(name.first);
			}

			return names;
		}

		bool Zip::directoryExists(const Util::String& Directory) const
		{
			return (zip.find(Directory) != zip.end());
		}

		void Zip::zipToFile(const Util::String& Filepath) const
		{
			auto data = zipToData();
			std::ofstream fout;
			fout.open(Filepath.getData(), std::ios::out | std::ios::trunc | std::ios::binary);

			if (!fout.good())
			{
				throw Util::Exception("Zip error, could not output file! [file=" + Filepath + ']', KRY_EXCEPTION_DATA);
			}
			else
			{
				fout.write(reinterpret_cast<const char*>(&data[0]), data.size());
			}

			fout.close();
		}

		std::vector<unsigned char> Zip::zipToData() const
		{
			std::vector<unsigned char> zipData;
			mz_zip_archive archive;
			memset(&archive, 0, sizeof(mz_zip_archive));

			if (!mz_zip_writer_init_heap(&archive, 0, 0))
			{
				throw Util::Exception("Zip error, couldn't initialize archive!", KRY_EXCEPTION_DATA);
			}

			for (auto& directory : zip)
			{
				Util::String folder;

				if (directory.first.getLength() > 0)
				{
					folder = directory.first;

					if (folder[folder.getLength() - 1] != '/')
						folder += '/';

					if (!mz_zip_writer_validate_archive_name(folder.getData()))
					{
						throw Util::Exception("Zip error, invalid directory name! [directory=" + folder + ']', KRY_EXCEPTION_DATA);
					}

					if (!mz_zip_writer_add_mem(&archive, folder.getData(), 0, 0, 0))
					{
						throw Util::Exception("Zip error, data write failed!", KRY_EXCEPTION_DATA);
					}
				}

				for (auto& file : directory.second.files)
				{
					Util::String filename = folder + file.first;
					auto& data = file.second;

					if (data.size() > 0)
					{
						if (!mz_zip_writer_validate_archive_name(filename.getData()))
						{
							throw Util::Exception("Zip error, invalid file name! [filename=" + filename + ']', KRY_EXCEPTION_DATA);
						}

						if (!mz_zip_writer_add_mem(&archive, filename.getData(), static_cast<const void*>(&data[0]), data.size(), MZ_BEST_COMPRESSION))
						{
							throw Util::Exception("Zip error, data write failed!", KRY_EXCEPTION_DATA);
						}
					}
				}
			}

			mz_zip_writer_finalize_archive(&archive);

			if (!archive.m_pState)
			{
				throw Util::Exception("Zip error, null archive!", KRY_EXCEPTION_DATA);
			}

			zipData.insert(zipData.begin(), static_cast<const unsigned char*>(archive.m_pState->m_pMem), static_cast<const unsigned char*>(archive.m_pState->m_pMem) + archive.m_pState->m_mem_size);
			mz_zip_writer_end(&archive);

			return zipData;
		}

		Zip::ZipDirectory& Zip::operator[](const Util::String& Directory)
		{
			auto i = zip.find(Directory);

			if (i == zip.end())
			{
				return zip.insert(std::pair<Util::String, ZipDirectory>(Directory, ZipDirectory())).first->second;
			}

			return i->second;
		}

		/* Zip::ZipDirectory */

		void Zip::ZipDirectory::clear()
		{
			files.clear();
		}

		std::vector<Util::String> Zip::ZipDirectory::getFileNames() const
		{
			std::vector<Util::String> names;

			for (auto& name : files)
			{
				names.push_back(name.first);
			}

			return names;
		}

		bool Zip::ZipDirectory::fileExists(const Util::String& File) const
		{
			return (files.find(File) != files.end());
		}

		std::vector<unsigned char>& Zip::ZipDirectory::operator[](const Util::String& File)
		{
			if (File.contains('/'))
			{
				throw Util::Exception("Zip error, illegal directory path symbol in file name!", KRY_EXCEPTION_DATA);
			}

			auto i = files.find(File);

			if (i == files.end())
			{
				return files.insert(std::pair<Util::String, std::vector<unsigned char> >(File, std::vector<unsigned char>())).first->second;
			}

			return i->second;
		}

	}
}