#include "Network\TCPServer.h"
#include "Network\TCPSocket.h"
#include "Utilities\BasicString.h"
#include "Utilities\ErrorMessage.h"
#include <ws2tcpip.h>
#include <Windows.h>

namespace kry
{
	namespace Network
	{
		TCPServer::TCPServer(unsigned short port) : socket(new TCPSocket(Util::String("0.0.0.0"), port)), maxconnections(SOMAXCONN), reuseaddress(true), listening(false)
		{
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to bind to port! " + Util::parseWinError(err) + " [port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
		}

		TCPServer::~TCPServer()
		{
			delete socket;
		}

		void TCPServer::bind(unsigned short port)
		{
			if (!socket->isClosed())
				socket->close();
			listening = false;
			socket->open(Util::String("0.0.0.0"), port);
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to bind to port! " + Util::parseWinError(err) + " [port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
		}

		void TCPServer::listen(unsigned backlog)
		{
			if (socket->getDescriptor() == INVALID_SOCKET)
				throw Util::Exception("Failed to begin listening! Server socket is invalid. [descriptor=" + Util::toString(socket->getDescriptor()) + ']', KRY_EXCEPTION_DATA);
			if (backlog != 0)
				maxconnections = backlog;
			if (::listen(socket->getDescriptor(), maxconnections) == SOCKET_ERROR)
				throw Util::Exception("Failed to begin listening! " + Util::parseWinError(WSAGetLastError()) + " [backlog=" + Util::toString(backlog) + ']', KRY_EXCEPTION_DATA);
			listening = true;
		}

		void TCPServer::stopListening()
		{
			socket->disconnect();
			listening = false;
		}

		bool TCPServer::accept()
		{
			if (!listening)
				return false;
			if (socket->getDescriptor() == INVALID_SOCKET)
				throw Util::Exception("Failed to attempt to accept! Server socket is invalid. [descriptor=" + Util::toString(socket->getDescriptor()) + ']', KRY_EXCEPTION_DATA);
			sockaddr_storage addr;
			int addrlen = sizeof(sockaddr_storage);
			SOCKET client = ::accept(socket->getDescriptor(), reinterpret_cast<sockaddr*>(&addr), &addrlen);
			if (client == INVALID_SOCKET)
			{
				if (listening) // if an error was thrown while the server was listening
				{
					addrlen = WSAGetLastError();
					socket->close();
					throw Util::Exception("Failed to accept connection! " + Util::parseWinError(addrlen) + " [address=" + socket->getAddress() + ", port=" + 
						Util::toString(socket->getPort()) + ']', KRY_EXCEPTION_DATA);
				}
				else
					return false; // check is done here because the server may have been closed while it was waiting
			}
			else if (addr.ss_family != AF_INET)
			{
				/** #TODO(incomplete) handle error message */
			}
			else if (addrlen != sizeof(sockaddr))
			{
				/** #TODO(incomplete) handle error message */
			}

			if (acceptevent)
				acceptevent(*this, std::shared_ptr<TCPSocket>(new TCPSocket(client, reinterpret_cast<SocketAddress>(&addr))));
			return true;
		}

		void TCPServer::close()
		{
			socket->close();
			listening = false;
		}

		void TCPServer::setReuseAddress(bool reuse)
		{
			reuseaddress = reuse;
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to re-bind to port! " + Util::parseWinError(err) + " [port=" + Util::toString(socket->getPort()) + ", reuseaddress=" + 
					Util::toString(reuse) + ']', KRY_EXCEPTION_DATA);
		}

		unsigned TCPServer::getBoundPort() const
		{
			return socket->getPort();
		}

		bool TCPServer::isClosed() const
		{
			return socket->isClosed();
		}

		int TCPServer::attemptBind()
		{
			SOCKET sock = socket->getDescriptor();
			int reuse = reuseaddress ? 1 : 0;
			if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, reinterpret_cast<char*>(&reuse), sizeof(reuse)) == SOCKET_ERROR)
			{
				reuse = WSAGetLastError();
				socket->disconnect();
				return reuse; // the calling function can throw
			}

			if (::bind(sock, socket->getAddressInfo()->ai_addr, socket->getAddressInfo()->ai_addrlen) == SOCKET_ERROR)
			{
				reuse = WSAGetLastError();
				socket->disconnect();
				return reuse;
			}
			return 0; // success
		}
	}
}
