#include "Network\TCPSocket.h"
#include "Network\Packet.h"
#include "Utilities\ErrorMessage.h"
#include "Utilities\Logger.h"
//#include <winsock2.h>
#include <ws2tcpip.h>
#include <Windows.h>

namespace kry
{
	namespace Network
	{
		TCPSocket::TCPSocket(const Util::String& address, unsigned short port) : address(address), info(nullptr), socket(INVALID_SOCKET), sendtimeout(30000), recvtimeout(30000), 
			port(port), closed(true), connected(false)
		{
			openFromDetails();
		}

		TCPSocket::TCPSocket(SocketDescriptor descriptor, const SocketAddress& sockaddress) : socket(descriptor), info(nullptr), sendtimeout(30000), recvtimeout(30000), 
			port(0), closed(true), connected(false)
		{
			if (socket != INVALID_SOCKET)
			{
				connected = true;

				sockaddr_in *adrs = sockaddress;
				char *tmp = inet_ntoa(adrs->sin_addr);
				address = std::move(Util::String(tmp, strlen(tmp)));
				port = adrs->sin_port;

				makeAddressInfo(address, Util::toString(port));

				closed = false;
			}
		}

		TCPSocket::TCPSocket(const TCPSocket& tcpsocket) : address(tcpsocket.address), info(nullptr), socket(INVALID_SOCKET), sendtimeout(tcpsocket.sendtimeout), 
			recvtimeout(tcpsocket.recvtimeout), port(tcpsocket.port), closed(true), connected(false)
		{
			openFromDetails();
		}

		TCPSocket::TCPSocket(TCPSocket&& tcpsocket) : address(tcpsocket.address), info(tcpsocket.info), socket(tcpsocket.socket), sendtimeout(tcpsocket.sendtimeout), 
			recvtimeout(tcpsocket.recvtimeout), port(tcpsocket.port), closed(tcpsocket.closed), connected(tcpsocket.connected)
		{
			tcpsocket.info = nullptr;
			tcpsocket.socket = INVALID_SOCKET;
			tcpsocket.closed = true;
		}

		TCPSocket::~TCPSocket()
		{
			try
			{
				destroy();
			}
			catch (const Util::Exception& e)
			{
				Util::Logger::log(e);
			}
		}

		void TCPSocket::open(const Util::String& address, unsigned short port)
		{
			if (!closed)
				close();
			this->address = address;
			this->port = port;
			openFromDetails();
		}

		void TCPSocket::connect()
		{
			if (socket == INVALID_SOCKET)
				throw Util::Exception("Failed to connect! Socket is invalid. [this=0x" + Util::toString(reinterpret_cast<unsigned>(this)) + ']', KRY_EXCEPTION_DATA);

			int err;
			if ((err = ::connect(socket, info->ai_addr, static_cast<int>(info->ai_addrlen))) == SOCKET_ERROR)
			{
				err = WSAGetLastError();
				freeaddrinfo(info);
				closesocket(socket);
				socket = INVALID_SOCKET;
				throw Util::Exception("Failed to connect! " + Util::parseWinError(err) + " [address=" + address + " , port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			}
			connected = true;
		}

		void TCPSocket::disconnect()
		{
			if (connected && !closed)
			{
				if (socket == INVALID_SOCKET)
					throw Util::Exception("Failed to disconnect! Socket has not been created. [this=0x" + Util::toString(reinterpret_cast<unsigned>(this)) + ']', KRY_EXCEPTION_DATA);
				int err;
				if ((err = shutdown(socket, SD_BOTH)) == SOCKET_ERROR)
				{
					err = WSAGetLastError();
					closesocket(socket);
					throw Util::Exception("Failed to disconnect! " + Util::parseWinError(err) + " [address=" + address + " , port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
				}
				connected = false;
			}
		}

		void TCPSocket::close()
		{
			if (connected)
				disconnect();
			if (socket != INVALID_SOCKET)
				if (closesocket(socket) == SOCKET_ERROR)
					throw Util::Exception("Failed to close the socket! " + Util::parseWinError(WSAGetLastError()) + " [address=" + address + " , port=" + 
						Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			socket = INVALID_SOCKET;
			closed = true;
		}

		void TCPSocket::send(const Packet& packet)
		{
			if (!connected)
				throw Util::Exception("Failed to send packet! Connection closed.", KRY_EXCEPTION_DATA);
			long long len = static_cast<long long>(packet.getSize() + packet.getHeaderSize());
			long sent = 0;
			const char *data = packet.getData();
			while (sent != SOCKET_ERROR && sent < len)
			{
				sent += ::send(socket, data + sent, static_cast<long>(len - sent), 0); // keep sending until all the data has been sent (the stream might send the data in chunks)
				if (sent == 0)
					break; // lost connection
			}

			if (sent == SOCKET_ERROR)
			{
				sent = WSAGetLastError();
				connected = false;
				close(); /** #TODO(change) not sure if the socket should be closed. test and research */
				throw Util::Exception("Failed to send packet! " + Util::parseWinError(sent) + " [address=" + address + " , port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			}
		}

		void TCPSocket::sendRaw(const char* data, unsigned length)
		{
			if (!connected)
				throw Util::Exception("Failed to send data! Connection closed.", KRY_EXCEPTION_DATA);
			long long len = static_cast<long long>(length);
			long sent = 0;
			while (sent != SOCKET_ERROR && sent < len)
			{
				sent += ::send(socket, data + sent, static_cast<long>(len - sent), 0); // keep sending until all the data has been sent (the stream might send the data in chunks)
				if (sent == 0)
					break; // lost connection
			}

			if (sent == SOCKET_ERROR)
			{
				sent = WSAGetLastError();
				connected = false;
				close(); /** #TODO(change) not sure if the socket should be closed. test and research */
				throw Util::Exception("Failed to send packet! " + Util::parseWinError(sent) + " [address=" + address + " , port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			}
		}

		void TCPSocket::receive(Packet& packet)
		{
			if (!connected)
				throw Util::Exception("Failed to begin receive! Connection closed.", KRY_EXCEPTION_DATA);
			long read = 0;
			long result = 0;
			char header[sizeof(unsigned)];
			if (recv(socket, header, sizeof(unsigned), MSG_PEEK) == SOCKET_ERROR)
			{
				/** #TODO(incomplete) handle error */
			}
			packet.resize(*(reinterpret_cast<unsigned*>(header)) - Packet::getHeaderSize());
			long long len = static_cast<long long>(packet.getSize() + Packet::getHeaderSize());
			while (result != SOCKET_ERROR && read < len)
			{
				result = ::recv(socket, packet.getData() + read, static_cast<long>(len - read), 0); // keep reading all data until all has been read
				if (result != SOCKET_ERROR)
				{
					read += result;
					if (result == 0 || !connected)
						break; // connection has been closed
				}
			}

			if (!connected)
				throw Util::Exception("Failed to receive packet! Connection closed.", KRY_EXCEPTION_DATA);
			if (result == SOCKET_ERROR)
			{
				connected = false;
				close();
				throw Util::Exception("Failed to receive packet! " + Util::parseWinError(WSAGetLastError()) + " [address=" + address + " , port=" + 
					Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			}
		}

		bool TCPSocket::receiveRaw(char* data, unsigned length)
		{
			if (!connected)
				throw Util::Exception("Failed to begin receive! Connection closed.", KRY_EXCEPTION_DATA);
			long long len = static_cast<long long>(length);
			long read = 0;
			long result = 0;
			while (result != SOCKET_ERROR && read < len)
			{
				result = ::recv(socket, data + read, static_cast<long>(len - read), 0);
				if (result != SOCKET_ERROR)
				{
					read += result;
					if (result == 0 || !connected)
						break; // connection has been closed or no more data left in the buffer
				}
			}

			if (!connected)
				throw Util::Exception("Failed to receive packet! Connection closed.", KRY_EXCEPTION_DATA);
			if (result == SOCKET_ERROR)
			{
				connected = false;
				close();
				throw Util::Exception("Failed to receive packet! " + Util::parseWinError(WSAGetLastError()) + " [address=" + address + " , port=" + 
					Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			}

			return read != 0;
		}

		void TCPSocket::setSendTimeout(unsigned timeout)
		{
			sendtimeout = timeout;
			if (socket != INVALID_SOCKET)
				setsockopt(socket, SOL_SOCKET, SO_SNDTIMEO, reinterpret_cast<const char*>(&sendtimeout), sizeof(sendtimeout));
		}

		void TCPSocket::setRecvTimeout(unsigned timeout)
		{
			recvtimeout = timeout;
			if (socket != INVALID_SOCKET)
				setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, reinterpret_cast<const char*>(&recvtimeout), sizeof(recvtimeout));
		}

		TCPSocket& TCPSocket::operator=(const TCPSocket& copy)
		{
			address = copy.address;
			sendtimeout = copy.sendtimeout;
			recvtimeout = copy.recvtimeout;
			port = copy.port;
			return *this;
		}

		TCPSocket& TCPSocket::operator=(TCPSocket&& move)
		{
			destroy();
			address = move.address;
			info = move.info;
			socket = move.socket;
			sendtimeout = move.sendtimeout;
			recvtimeout = move.recvtimeout;
			port = move.port;
			closed = move.closed;
			connected = move.connected;
			
			move.info = nullptr;
			move.socket = INVALID_SOCKET;
			move.closed = true;
			move.connected = false;
			return *this;
		}

		void TCPSocket::openFromDetails()
		{
			Util::String p = Util::toString(port);
			makeAddressInfo(address, p);

			socket = ::socket(info->ai_family, info->ai_socktype, info->ai_protocol);
			if (socket == INVALID_SOCKET)
			{
				freeaddrinfo(info);
				throw Util::Exception("Failed to initialise socket! " + Util::parseWinError(WSAGetLastError()) + " [address=" + address + ", port=" + p + ']', KRY_EXCEPTION_DATA);
			}
			//freeaddrinfo(info); /** #TODO(change) why did i free address info here? investigate */

			closed = false;
		}

		void TCPSocket::makeAddressInfo(const Util::String& addr, const Util::String& prt)
		{
			addrinfo *ifo = nullptr, hint{0};

			hint.ai_family = AF_UNSPEC;
			hint.ai_socktype = SOCK_STREAM;
			hint.ai_protocol = IPPROTO_TCP;

			int err;
			if ((err = getaddrinfo(addr.isEmpty() ? 0 : addr.getData(), prt.getData(), &hint, &ifo)) != 0)
			{
				if (ifo != nullptr)
					freeaddrinfo(ifo);
				throw Util::Exception("Failed to get address info! " + Util::parseWinError(err) + " [address=" + address + ", port=" + prt + ']', KRY_EXCEPTION_DATA);
			}

			info = ifo;
		}

		void TCPSocket::destroy()
		{
			if (info != nullptr)
			{
				freeaddrinfo(info);
				info = nullptr;
			}
			close();
		}
	}
}
