#include "Network\UDPSocket.h"
#include "Network\DatagramPacket.h"
#include "Utilities\ErrorMessage.h"
#include "Utilities\Logger.h"
#include <ws2tcpip.h>
#include <Windows.h>

namespace kry
{
	namespace Network
	{
		const Util::String UDPSocket::ANY_ADDRESS("0.0.0.0");

		UDPSocket::UDPSocket(unsigned short port, const Util::String& localaddress, unsigned sendtimeout, unsigned recvtimeout) : address(nullptr), socket(INVALID_SOCKET), sendtimeout(sendtimeout), 
			recvtimeout(recvtimeout), closed(true), reuseaddress(true)
		{
			openSocket();

			sockaddr_in *addr = reinterpret_cast<sockaddr_in*>(malloc(sizeof(sockaddr_storage)));
			ZeroMemory(addr, sizeof(sockaddr_storage));
			reinterpret_cast<sockaddr_in*>(addr)->sin_family = AF_INET;
			reinterpret_cast<sockaddr_in*>(addr)->sin_addr.S_un.S_addr = inet_addr(localaddress.getData()); 
			reinterpret_cast<sockaddr_in*>(addr)->sin_port = port;
			address = reinterpret_cast<SocketAddress>(addr);

			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to bind socket! " + Util::parseWinError(err) + " [port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
		}

		UDPSocket::UDPSocket(const UDPSocket& copy) : address(nullptr), socket(INVALID_SOCKET), sendtimeout(copy.sendtimeout), recvtimeout(copy.recvtimeout), closed(true), 
			reuseaddress(copy.reuseaddress)
		{
			openSocket();

			address = reinterpret_cast<SocketAddress>(malloc(sizeof(sockaddr_storage)));
			memcpy(address, copy.address, sizeof(sockaddr_storage));
			
			int err; /** #TODO(test) not sure if this socket can bind to the same address as the copied socket */
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to bind socket! " + Util::parseWinError(err) + " [port=" + Util::toString(copy.getBoundPort()) + ']', KRY_EXCEPTION_DATA);
		}

		UDPSocket::UDPSocket(UDPSocket&& move) : address(move.address), socket(move.socket), sendtimeout(move.sendtimeout), recvtimeout(move.recvtimeout), 
			closed(move.closed), reuseaddress(move.reuseaddress)
		{
			move.address = nullptr;
			move.socket = INVALID_SOCKET;
		}

		UDPSocket::~UDPSocket()
		{
			free(reinterpret_cast<void*>(address)); // using malloc, so use free instead of delete
			try
			{
				close();
			}
			catch (const Util::Exception& e)
			{
				Util::Logger::log(e);
			}
		}

		void UDPSocket::bind(unsigned short port, const Util::String& localaddress)
		{
			if (address == nullptr)
				throw Util::Exception("Failed to bind socket! Socket address is invalid.", KRY_EXCEPTION_DATA);
			
			unsigned long inet = inet_addr(localaddress.getData());
			if (!closed)
			{
				close();
				openSocket();
			}
			else if (address->sin_port != port || address->sin_addr.S_un.S_addr != inet)
				openSocket();
			
			address->sin_addr.S_un.S_addr = inet; 
			address->sin_port = port;
			
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to bind socket! " + Util::parseWinError(err) + " [port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
		}

		void UDPSocket::send(const DatagramPacket& packet)
		{
			send(packet, packet.getAddress(), packet.getPort());
		}

		void UDPSocket::send(const Packet& packet, const Util::String& ipaddress, unsigned short port)
		{
			if (closed)
				throw Util::Exception("Failed to send packet! Socket is closed.", KRY_EXCEPTION_DATA);

			sockaddr_storage dest;
			reinterpret_cast<sockaddr_in*>(&dest)->sin_family = AF_INET;
			reinterpret_cast<sockaddr_in*>(&dest)->sin_addr.S_un.S_addr = inet_addr(ipaddress.getData());
			reinterpret_cast<sockaddr_in*>(&dest)->sin_port = port;

			long long len = static_cast<long long>(packet.getSize() + packet.getHeaderSize());
			long sent = 0;
			const char* data = packet.getData();
			while (sent != SOCKET_ERROR && sent < len)
			{
				sent += ::sendto(socket, data + sent, static_cast<long>(len - sent), 0, reinterpret_cast<sockaddr*>(&dest), sizeof(sockaddr_storage));
				if (sent == 0)
					break; // lost connection
			}

			if (sent == SOCKET_ERROR)
			{
				sent = WSAGetLastError();
				close(); // not sure if the socket should be closed
				throw Util::Exception("Failed to send packet! " + Util::parseWinError(sent) + " [address=" + ipaddress + " , port=" + Util::toString(port) + ']', KRY_EXCEPTION_DATA);
			}
		}

		bool UDPSocket::receive(DatagramPacket& packet)
		{
			if (closed)
				throw Util::Exception("Failed to receive packet! Socket is closed.", KRY_EXCEPTION_DATA);

			sockaddr_storage from;
			int fromlen = sizeof(from);

			long read = 0;
			char header[sizeof(unsigned)];
			if (recvfrom(socket, header, sizeof(unsigned), MSG_PEEK, (sockaddr*)&from, &fromlen) == SOCKET_ERROR)
				if (WSAGetLastError() != WSAEMSGSIZE)
					throw Util::Exception("Failed to receive! " + Util::parseWinError(WSAGetLastError()), KRY_EXCEPTION_DATA);

			packet.resize(*(reinterpret_cast<unsigned*>(header)) - Packet::getHeaderSize());
			char* addr = inet_ntoa(reinterpret_cast<sockaddr_in*>(&from)->sin_addr);
			packet.setAddress(Util::String(addr, strlen(addr))); // set the address/port for the packet
			packet.setPort(reinterpret_cast<sockaddr_in*>(&from)->sin_port);

			long long len = static_cast<long long>(packet.getSize() + packet.getHeaderSize());
			while (read != SOCKET_ERROR && read < len)
			{
				read += ::recvfrom(socket, packet.getData() + read, static_cast<long>(len - read), 0, reinterpret_cast<sockaddr*>(&from), &fromlen);
				if (read == 0 || closed)
					return false; // connection has been closed
			}

			if (read == SOCKET_ERROR)
			{
				close();
				throw Util::Exception("Failed to receive packet! " + Util::parseWinError(WSAGetLastError()) + " [address=" + packet.getAddress() + " , port=" + 
					Util::toString(packet.getPort()) + ']', KRY_EXCEPTION_DATA);
			}
			return !closed;
		}

		void UDPSocket::close()
		{
			if (socket != INVALID_SOCKET)
				if (closesocket(socket) == SOCKET_ERROR)
					throw Util::Exception("Failed to close the datagram socket! " + Util::parseWinError(WSAGetLastError()), KRY_EXCEPTION_DATA);
			socket = INVALID_SOCKET;
			closed = true;
		}

		Util::String UDPSocket::getBoundAddress() const
		{
			if (address == nullptr)
				throw Util::Exception("Invalid socket address!", KRY_EXCEPTION_DATA);
			char* addr = inet_ntoa(address->sin_addr);
			return Util::String(addr, strlen(addr));
		}

		unsigned short UDPSocket::getBoundPort() const
		{
			if (address == nullptr)
				throw Util::Exception("Invalid socket address!", KRY_EXCEPTION_DATA);
			return address->sin_port;
		}

		void UDPSocket::setReuseAddress(bool reuse)
		{
			reuseaddress = reuse;
			if (!closed)
				close();
			openSocket();
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to re-bind socket! " + Util::parseWinError(err) + " [port=" + Util::toString(address->sin_port) + ", reuseaddress=" + 
					Util::toString(reuseaddress) + ']', KRY_EXCEPTION_DATA);
		}

		void UDPSocket::setSendTimeout(unsigned timeout)
		{
			sendtimeout = timeout;
			if (!closed)
				close();
			openSocket();
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to re-bind socket! " + Util::parseWinError(err) + " [port=" + Util::toString(address->sin_port) + ", sendtimeout=" + 
					Util::toString(sendtimeout) + ']', KRY_EXCEPTION_DATA);
		}

		void UDPSocket::setRecvTimeout(unsigned timeout)
		{
			recvtimeout = timeout;
			if (!closed)
				close();
			openSocket();
			int err;
			if ((err = attemptBind()) != 0)
				throw Util::Exception("Failed to re-bind socket! " + Util::parseWinError(err) + " [port=" + Util::toString(address->sin_port) + ", recvtimeout=" + 
					Util::toString(recvtimeout) + ']', KRY_EXCEPTION_DATA);
		}

		void UDPSocket::openSocket()
		{
			socket = ::socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
			if (socket == INVALID_SOCKET)
				throw Util::Exception("Failed to create datagram socket! " + Util::parseWinError(WSAGetLastError()), KRY_EXCEPTION_DATA);
			closed = false;
		}

		UDPSocket& UDPSocket::operator=(const UDPSocket& copy)
		{
			memcpy(address, copy.address, sizeof(sockaddr_storage));
			sendtimeout = copy.sendtimeout;
			recvtimeout = copy.recvtimeout;
			reuseaddress = copy.reuseaddress;
			return *this;
		}

		UDPSocket& UDPSocket::operator=(UDPSocket&& move)
		{
			free(reinterpret_cast<void*>(address)); // using malloc, so use free instead of delete
			close();

			address = move.address;
			socket = move.socket;
			sendtimeout = move.sendtimeout;
			recvtimeout = move.recvtimeout;
			closed = move.closed;
			reuseaddress = move.reuseaddress;
			
			move.address = nullptr;
			move.socket = INVALID_SOCKET;
			move.closed = true;
			return *this;
		}

		int UDPSocket::attemptBind()
		{
			int reuse = reuseaddress ? 1 : 0;
			if (setsockopt(socket, SOL_SOCKET, SO_REUSEADDR, reinterpret_cast<char*>(&reuse), sizeof(reuse)) == SOCKET_ERROR)
				return WSAGetLastError(); // the calling function can throw

			if (setsockopt(socket, SOL_SOCKET, SO_SNDTIMEO, reinterpret_cast<char*>(&sendtimeout), sizeof(sendtimeout)) == SOCKET_ERROR)
				return WSAGetLastError(); // the calling function can throw

			if (setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, reinterpret_cast<char*>(&recvtimeout), sizeof(recvtimeout)) == SOCKET_ERROR)
				return WSAGetLastError(); // the calling function can throw

			if (::bind(socket, reinterpret_cast<sockaddr*>(address), sizeof(sockaddr_storage)) == SOCKET_ERROR)
				return WSAGetLastError();

			return 0; // success
		}
	}
}
