#include "System\Clipboard.h"
#include "Utilities\BasicString.h"
#include "Utilities\Exception.h"
#include "Utilities\ErrorMessage.h"
#include <Windows.h>

namespace kry
{
	namespace System
	{
		void openClip()
		{
			if (OpenClipboard(0) == FALSE)
				throw Util::Exception("Failed to open clipboard! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}

		void closeClip()
		{
			if (CloseClipboard() == FALSE)
				throw Util::Exception("Failed to close clipboard! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}

		void* lockClip(HANDLE handle)
		{
			void* result = GlobalLock(handle);
			if (result == nullptr)
				throw Util::Exception("Failed to lock clipboard! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			return result;
		}

		void unlockClip(HANDLE handle)
		{
			if (GlobalUnlock(handle) == FALSE && GetLastError() != NO_ERROR)
				throw Util::Exception("Failed to unlock clipboard! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}


		void writeCliboardData(const Util::String& formatName, const char* data, unsigned size)
		{
			unsigned format = RegisterClipboardFormat(formatName.getData());
			openClip();
			HGLOBAL buffer = GlobalAlloc(GMEM_DDESHARE, size);
			void* ptr = lockClip(buffer);
			memcpy(ptr, data, size);
			unlockClip(buffer);
			if (SetClipboardData(format, buffer) == FALSE)
				throw Util::Exception("Failed to set clipboard data! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			closeClip();
		}

		void writeClipboardText(const Util::String& data)
		{
			HANDLE handle = GlobalAlloc(GMEM_DDESHARE, data.getLength() + 1);
			if (handle == nullptr)
				throw Util::Exception("Failed to allocate memory for clipboard data! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);

			void* ptr = lockClip(handle);
			memcpy(ptr, data.getData(), data.getLength() + 1);
			unlockClip(handle);
			clearClipboard();
			openClip();
			if (SetClipboardData(CF_TEXT, handle) == FALSE)
				throw Util::Exception("Failed to set clipboard text! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			closeClip();
		}

		std::vector<char> getClipboardData(const Util::String& formatName, unsigned size)
		{
			unsigned format = RegisterClipboardFormat(formatName.getData());
			openClip();
			HANDLE handle = GetClipboardData(format);
			char* data = static_cast<char*>(lockClip(handle));
			std::vector<char> buffer(size);
			for (unsigned i = 0; i < size; ++i)
				buffer.push_back(data[i]);
			unlockClip(handle);
			closeClip();
			return buffer;
		}

		Util::String getClipboardText()
		{
			openClip();
			HANDLE handle = GetClipboardData(CF_TEXT);
			if (handle == nullptr)
				throw Util::Exception("Failed to get text from clipboard! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);

			char* data = static_cast<char*>(lockClip(handle));
			unsigned size = static_cast<unsigned>(GlobalSize(handle));
			unlockClip(handle);
			closeClip();
			return Util::String(data, size - 1);
		}

		void clearClipboard()
		{
			openClip();
			if (EmptyClipboard() == FALSE)
				throw Util::Exception("Failed to empty clipboard! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			closeClip();
		}
	}
}
