#include "System\Filesystem.h"
#include "Utilities\BasicString.h"
#include "Utilities\ErrorMessage.h"
#include <Windows.h>
#include <fstream>
#include <string> // bit grub lol, but its quickers

namespace kry
{
	namespace System
	{
		bool isFile(const Util::String& path) // weak check to see if a path is a file and not a folder
		{
			return path[path.getLength() - 1] != '\\' && path[path.getLength() - 1] != '*';
		}

		bool checkLength(const Util::String& path)
		{
			return path.getLength() < MAX_PATH;
		}


		
		FileAccess operator|(const FileAccess left, const FileAccess right)
		{
			typedef std::underlying_type<FileAccess>::type UnType;
			return static_cast<FileAccess>(static_cast<UnType>(left) | static_cast<UnType>(right));
		}

		FileAccess operator&(const FileAccess left, const FileAccess right)
		{
			typedef std::underlying_type<FileAccess>::type UnType;
			return static_cast<FileAccess>(static_cast<UnType>(left) & static_cast<UnType>(right));
		}

		bool fileExists(const Util::String& filename)
		{
			if (!checkLength(filename))
				throw Util::Exception("Invalid file name! File name exceeds max length. [pathlength=" + Util::toString(filename.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			WIN32_FIND_DATA data;
			HANDLE handle = FindFirstFile(filename.getData(), &data);
			if (handle == INVALID_HANDLE_VALUE)
				return false;//throw Util::Exception("Invalid file name! " + Util::parseWinError(GetLastError()) + " [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			if (handle != nullptr)
				FindClose(handle);

			return reinterpret_cast<long>(handle) != ERROR_FILE_NOT_FOUND;
		}

		bool pathExists(const Util::String& directory)
		{
			if (!checkLength(directory))
				throw Util::Exception("Invalid path name! Path name exceeds max length. [pathlength=" + Util::toString(directory.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(directory))
				throw Util::Exception("Invalid path name! Must not end with a \\ or *. [pathname=" + directory + ']', KRY_EXCEPTION_DATA);
			WIN32_FIND_DATA data;
			HANDLE handle = FindFirstFile(directory.getData(), &data);
			if (handle == INVALID_HANDLE_VALUE)
				return false;//throw Util::Exception("Invalid path name! " + Util::parseWinError(GetLastError()) + " [pathname=" + directory + ']', KRY_EXCEPTION_DATA);
			if (handle != nullptr)
				FindClose(handle);

			return reinterpret_cast<long>(handle) != ERROR_FILE_NOT_FOUND;
		}

		void createFile(const Util::String& filename, bool overwrite, FileAccess access)
		{
			if (!checkLength(filename))
				throw Util::Exception("Invalid file name! File name exceeds max length. [pathlength=" + Util::toString(filename.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			DWORD dwaccess = 0;
			if (access != FileAccess::NONE)
			{
				if ((access & FileAccess::READ) == FileAccess::READ)
					dwaccess |= GENERIC_READ;
				if ((access & FileAccess::WRITE) == FileAccess::WRITE)
					dwaccess |= GENERIC_WRITE;
				if ((access & FileAccess::EXECUTE) == FileAccess::EXECUTE)
					dwaccess |= GENERIC_EXECUTE;
			}
			HANDLE handle = CreateFile(filename.getData(), dwaccess, FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE, nullptr, overwrite ? CREATE_NEW : OPEN_ALWAYS, 
										FILE_ATTRIBUTE_NORMAL, nullptr);
			if (handle == INVALID_HANDLE_VALUE || handle == nullptr)
			{
				if (GetLastError() == ERROR_FILE_EXISTS)
					throw Util::Exception("Failed to create file! File already exists. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
				throw Util::Exception("Failed to create file! " + Util::parseWinError(GetLastError()) + " [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			}
			CloseHandle(handle);
		}

		void createPath(const Util::String& directory)
		{
			if (!checkLength(directory))
				throw Util::Exception("Invalid path name! Path name exceeds max length. [pathlength=" + Util::toString(directory.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(directory))
				throw Util::Exception("Invalid path name! Must not end with a \\ or *. [pathname=" + directory + ']', KRY_EXCEPTION_DATA);
			if (CreateDirectory(directory.getData(), nullptr) == FALSE)
			{
				if (GetLastError() == ERROR_ALREADY_EXISTS)
					throw Util::Exception("Failed to create directory! Path already exists. [pathname=" + directory + ']', KRY_EXCEPTION_DATA);
				throw Util::Exception("Failed to create directory! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}
		}

		void copyFile(const Util::String& source, const Util::String& destination)
		{
			if (!checkLength(source))
				throw Util::Exception("Invalid source name! Source name exceeds max length. [sourcelength=" + Util::toString(source.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!checkLength(destination))
				throw Util::Exception("Invalid destination name! Destination name exceeds max length. [destname=" + Util::toString(destination.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(source))
				throw Util::Exception("Invalid source name! Must not end with a \\ or *. [sourcename=" + source + ']', KRY_EXCEPTION_DATA);
			if (!isFile(destination))
				throw Util::Exception("Invalid destination name! Must not end with a \\ or *. [destname=" + source + ']', KRY_EXCEPTION_DATA);
			if (CopyFile(source.getData(), destination.getData(), TRUE) == FALSE)
			{
				if (GetLastError() == ERROR_ACCESS_DENIED)
					throw Util::Exception("Failed to copy file! Access denied. [source=" + source + ", destination=" + destination + ']', KRY_EXCEPTION_DATA);
				throw Util::Exception("Failed to copy file! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}
		}
		
		void copyPath(const Util::String& source, const Util::String& destination)
		{
			/** #TODO(incomplete) need to finish this without SH operations, preferably. could get fat. bit of research needed */
		}

		void moveFile(const Util::String& source, const Util::String& destination, bool waitForCompletion)
		{
			if (!checkLength(source))
				throw Util::Exception("Invalid source name! Source name exceeds max length. [sourcelength=" + Util::toString(source.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!checkLength(destination))
				throw Util::Exception("Invalid destination name! Destination name exceeds max length. [destname=" + Util::toString(destination.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(source))
				throw Util::Exception("Invalid source name! Must not end with a \\ or *. [sourcename=" + source + ']', KRY_EXCEPTION_DATA);
			if (!isFile(destination))
				throw Util::Exception("Invalid destination name! Must not end with a \\ or *. [destname=" + source + ']', KRY_EXCEPTION_DATA);
			if (MoveFileEx(source.getData(), destination.getData(), MOVEFILE_COPY_ALLOWED | (waitForCompletion ? MOVEFILE_WRITE_THROUGH : 0)) == FALSE)
				throw Util::Exception("Failed to move file! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}

		void movePath(const Util::String& source, const Util::String& destination)
		{
			/** #TODO(incomplete) same sitch as copyPath */
		}

		void deleteFile(const Util::String& filename)
		{
			if (!checkLength(filename))
				throw Util::Exception("Invalid file name! File name exceeds max length. [pathlength=" + Util::toString(filename.getLength()) + ']', KRY_EXCEPTION_DATA);
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			if (DeleteFile(filename.getData()) == FALSE)
				throw Util::Exception("Failed to delete file! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}

		void deletePath(const Util::String& filename)
		{
			/** #TODO(incomplete) once again. MS recommends using SHFileOperation for such methods, might look into it */
		}

		void writeAllLines(const Util::String& filename, const Util::String& lines, char delimiter, FileMode mode)
		{
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			std::ofstream file(filename.getData(), (mode == FileMode::APPEND ? std::ios::app : std::ios::out));
			if (!file)
				throw Util::Exception("Failed to open file! [filename=" + filename + ']', KRY_EXCEPTION_DATA);

			std::vector<Util::String> vlines(std::move(lines.explode(delimiter)));
			for (auto& line : vlines)
				file << line << '\n';
			file.close();
		}

		void writeAllLines(const Util::String& filename, const std::vector<Util::String>& lines, FileMode mode)
		{
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			std::ofstream file(filename.getData(), (mode == FileMode::APPEND ? std::ios::app : std::ios::out));
			if (!file)
				throw Util::Exception("Failed to open file for writing! [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			
			for (auto& line : lines)
				file << line << '\n';
			file.close();
		}

		std::vector<Util::String> readAllLines(const Util::String& filename, char delimiter)
		{
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			std::ifstream file(filename.getData());
			if (!file)
				throw Util::Exception("Failed to open file for reading! [filename=" + filename + ']', KRY_EXCEPTION_DATA);

			std::vector<Util::String> lines;
			std::string line;
			while (std::getline(file, line, delimiter))
				if (!line.empty())
					lines.push_back(Util::String(line.data(), line.length()));
			file.close();

			return lines;
		}

		void writeAllBytes(const Util::String& filename, const char* buffer, unsigned bufferlength, FileMode mode)
		{
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			std::ofstream file(filename.getData(), (mode == FileMode::APPEND ? std::ios::app : std::ios::out));
			if (!file)
				throw Util::Exception("Failed to open file for writing! [filename=" + filename + ']', KRY_EXCEPTION_DATA);

			file.write(buffer, bufferlength);
			file.close();
		}

		void writeAllBytes(const Util::String& filename, const std::vector<char>& buffer, FileMode mode)
		{
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			std::ofstream file(filename.getData(), (mode == FileMode::APPEND ? std::ios::app : std::ios::out));
			if (!file)
				throw Util::Exception("Failed to open file for writing! [filename=" + filename + ']', KRY_EXCEPTION_DATA);

			file.write(buffer.data(), buffer.size());
			file.close();
		}

		std::vector<char> readAllBytes(const Util::String& filename)
		{
			if (!isFile(filename))
				throw Util::Exception("Invalid file name! Must not end with a \\ or *. [filename=" + filename + ']', KRY_EXCEPTION_DATA);
			std::ifstream file(filename.getData());
			if (!file)
				throw Util::Exception("Failed to open file for reading! [filename=" + filename + ']', KRY_EXCEPTION_DATA);

			file.seekg(0, std::ios::end);
			unsigned size = static_cast<unsigned>(file.tellg()); // get size of file
			file.seekg(std::ios::beg);
			file.clear(std::ios::eofbit);

			std::vector<char> bytes(size); // set size and capacity to size of file
			bytes.resize(size);
			file.read(bytes.data(), size); // read all
			file.close();

			return bytes;
		}

		std::vector<Util::String> getAllFiles(const Util::String& directory, bool shortened)
		{
			if (!pathExists(directory))
				throw Util::Exception("Failed to get filenames! Path does not exist. [pathname=" + directory + ']', KRY_EXCEPTION_DATA);
			WIN32_FIND_DATA data{0};
			HANDLE handle = FindFirstFile((directory + "\\*").getData(), &data);

			std::vector<Util::String> files;
			do
			{
				Util::String file(data.cFileName);
				files.push_back((shortened ? "" : directory + '\\') + file.substring(0, file.indexOf('\0'))); /** #TODO(test) indexOf part might fail. test for a length of 259 */
			} while (FindNextFile(handle, &data) != FALSE);

			auto error = GetLastError();
			if (error != ERROR_NO_MORE_FILES)
			{
				FindClose(handle);
				throw Util::Exception("Failed to get file names! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}

			if (FindClose(handle) == FALSE)
				throw Util::Exception("Failed to close find handle! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);

			return files;
		}

		std::vector<Util::String> getAllFiles(const Util::String& directory, const std::vector<Util::String>& extfilter, bool shortened)
		{
			if (!pathExists(directory))
				throw Util::Exception("Failed to get filenames! Path does not exist. [pathname=" + directory + ']', KRY_EXCEPTION_DATA);
			WIN32_FIND_DATA data{0};
			HANDLE handle = FindFirstFile((directory + "\\*").getData(), &data);

			std::vector<Util::String> files;
			do
			{
				Util::String file(data.cFileName);
				bool filtered = false;
				for (auto& ext : extfilter)
				{
					if (file.endsWith(ext))
					{
						filtered = true;
						break;
					}
				}
				if (filtered)
					continue;

				files.push_back((shortened ? "" : directory + '\\') + file.substring(0, file.indexOf('\0'))); /** #TODO(test) indexOf part might fail. test for a length of 259 */
			} while (FindNextFile(handle, &data) != FALSE);

			auto error = GetLastError();
			if (error != ERROR_NO_MORE_FILES)
			{
				FindClose(handle);
				throw Util::Exception("Failed to get file names! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}

			if (FindClose(handle) == FALSE)
				throw Util::Exception("Failed to close find handle! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);

			return files;
		}

		std::vector<Util::String> getAllPaths(const Util::String& directory, bool shortened)
		{
			/** #TODO(incomplete) not sure if this is the same as getAllFiles or not. needs testing */
			return std::vector<Util::String>(); // dud
		}
	}
}
