#include "System\Window.h"
#include "Utilities\BasicString.h"
#include "Utilities\BasicVector.h"
#include "Utilities\ErrorMessage.h"
#include "Utilities\Logger.h"
#include <cmath>
#include <Windows.h>
#include <GL\glew.h>
#include <GL\wglew.h>

namespace kry
{
	namespace System
	{
		static bool keystates[static_cast<unsigned>(Key::F15)]{false}; /** #TODO(refactor) global forbidden by code conventions, maybe move it to Input.h or something */
		static bool buttonstates[static_cast<unsigned>(Button::X2)]{false};

		Key virtualKeyToKey(unsigned key, long flags) // same with this
		{
			switch (key)
			{
				case VK_SHIFT:
				{
					static UINT lshift = MapVirtualKey(VK_LSHIFT, MAPVK_VK_TO_VSC);
					UINT scancode = static_cast<UINT>((flags & (0xFF << 16)) >> 16);
					return scancode == lshift ? Key::SHIFT_L : Key::SHIFT_R;
				}
				case VK_MENU:		return (HIWORD(flags) & KF_EXTENDED) != 0 ? Key::ALT_R : Key::ALT_L;
				case VK_CONTROL:	return (HIWORD(flags) & KF_EXTENDED) != 0 ? Key::CONTROL_R : Key::CONTROL_L;

				case 'A':			return Key::A;
				case 'B':			return Key::B;
				case 'C':			return Key::C;
				case 'D':			return Key::D;
				case 'E':			return Key::E;
				case 'F':			return Key::F;
				case 'G':			return Key::G;
				case 'H':			return Key::H;
				case 'I':			return Key::I;
				case 'J':			return Key::J;
				case 'K':			return Key::K;
				case 'L':			return Key::L;
				case 'M':			return Key::M;
				case 'N':			return Key::N;
				case 'O':			return Key::O;
				case 'P':			return Key::P;
				case 'Q':			return Key::Q;
				case 'R':			return Key::R;
				case 'S':			return Key::S;
				case 'T':			return Key::T;
				case 'U':			return Key::U;
				case 'V':			return Key::V;
				case 'W':			return Key::W;
				case 'X':			return Key::X;
				case 'Y':			return Key::Y;
				case 'Z':			return Key::Z;
				case '0':			return Key::NUMBER_0;
				case '1':			return Key::NUMBER_1;
				case '2':			return Key::NUMBER_2;
				case '3':			return Key::NUMBER_3;
				case '4':			return Key::NUMBER_4;
				case '5':			return Key::NUMBER_5;
				case '6':			return Key::NUMBER_6;
				case '7':			return Key::NUMBER_7;
				case '8':			return Key::NUMBER_8;
				case '9':			return Key::NUMBER_9;
				case VK_NUMPAD0:	return Key::NUMPAD_0;
				case VK_NUMPAD1:	return Key::NUMPAD_1;
				case VK_NUMPAD2:	return Key::NUMPAD_2;
				case VK_NUMPAD3:	return Key::NUMPAD_3;
				case VK_NUMPAD4:	return Key::NUMPAD_4;
				case VK_NUMPAD5:	return Key::NUMPAD_5;
				case VK_NUMPAD6:	return Key::NUMPAD_6;
				case VK_NUMPAD7:	return Key::NUMPAD_7;
				case VK_NUMPAD8:	return Key::NUMPAD_8;
				case VK_NUMPAD9:	return Key::NUMPAD_9;
				case VK_NUMLOCK:	return Key::NUMLOCK;
				case VK_ADD:		return Key::ADD;
				case VK_SUBTRACT:	return Key::SUBSTRACT;
				case VK_MULTIPLY:	return Key::MULTIPLY;
				case VK_DIVIDE:		return Key::DIVIDE;
				case VK_DECIMAL:	return Key::DECIMAL;
				case VK_OEM_COMMA:	return Key::COMMA;
				case VK_OEM_PERIOD:	return Key::PERIOD;
				case VK_OEM_2:		return Key::SLASH_FORWARD;
				case VK_OEM_5:		return Key::SLASH_BACKWARD;
				case VK_OEM_1:		return Key::SEMICOLON;
				case VK_OEM_7:		return Key::QUOTE;
				case VK_OEM_4:		return Key::BRACKET_LEFT;
				case VK_OEM_6:		return Key::BRACKET_RIGHT;
				case VK_OEM_3:		return Key::TILDE;
				case VK_OEM_MINUS:	return Key::DASH;
				case VK_OEM_PLUS:	return Key::EQUALS;
				case VK_UP:			return Key::UP;
				case VK_DOWN:		return Key::DOWN;
				case VK_LEFT:		return Key::LEFT;
				case VK_RIGHT:		return Key::RIGHT;
				case VK_INSERT:		return Key::INSERT;
#if defined(DELETE)
	#define DELETE_PUSHED
	#pragma push_macro("DELETE")
	#undef DELETE
#endif
				case VK_DELETE:		return Key::DELETE;
#if defined(DELETE_PUSHED)
	#undef DELETE_PUSHED
	#pragma pop_macro("DELETE")
#endif
				case VK_HOME:		return Key::HOME;
				case VK_END:		return Key::END;
				case VK_PRIOR:		return Key::PAGE_UP;
				case VK_NEXT:		return Key::PAGE_DOWN;
				case VK_BACK:		return Key::BACKSPACE;
				case VK_SPACE:		return Key::SPACE;
				case VK_ESCAPE:		return Key::ESCAPE;
				case VK_RETURN:		return Key::RETURN;
				case VK_SNAPSHOT:	return Key::PRINT_SCREEN;
				case VK_SCROLL:		return Key::SCROLL_LOCK;
				case VK_PAUSE:		return Key::PAUSE;
				case VK_TAB:		return Key::TAB;
				case VK_CAPITAL:	return Key::CAPS_LOCK;
				case VK_RCONTROL:	return Key::CONTROL_R;
				case VK_LCONTROL:	return Key::CONTROL_L;
				case VK_RSHIFT:		return Key::SHIFT_R;
				case VK_LSHIFT:		return Key::SHIFT_L;
				case VK_RMENU:		return Key::ALT_R;
				case VK_LMENU:		return Key::ALT_L;
				case VK_RWIN:		return Key::SYSTEM_R;
				case VK_LWIN:		return Key::SYSTEM_L;
				case VK_APPS:		return Key::MENU;
				case VK_F1:			return Key::F1;
				case VK_F2:			return Key::F2;
				case VK_F3:			return Key::F3;
				case VK_F4:			return Key::F4;
				case VK_F5:			return Key::F5;
				case VK_F6:			return Key::F6;
				case VK_F7:			return Key::F7;
				case VK_F8:			return Key::F8;
				case VK_F9:			return Key::F9;
				case VK_F10:		return Key::F10;
				case VK_F11:		return Key::F11;
				case VK_F12:		return Key::F12;
				case VK_F13:		return Key::F13;
				case VK_F14:		return Key::F14;
				case VK_F15:		return Key::F15;
			}
			return Key::UNKNOWN;
		}


		WindowClass Window::winclass(nullptr);
		unsigned Window::wincount(0);

		Window::Window(const Util::String& title, const Util::Vector2i& size) : handle(nullptr), context(nullptr), glcontext(nullptr), destroyed(true), 
			winmode(WindowMode::WINDOWED)
		{
			if (winclass == nullptr)
			{
				winclass = new tagWNDCLASSEXA{0};

				winclass->cbSize = sizeof(tagWNDCLASSEXA);
				winclass->style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
				winclass->lpfnWndProc = winProc;
				winclass->hInstance = GetModuleHandle(nullptr);
				winclass->hIcon = LoadIcon(0, IDI_APPLICATION);
				winclass->hCursor = LoadCursor(0, IDC_ARROW);
				winclass->hbrBackground = reinterpret_cast<HBRUSH>(GetStockObject(BLACK_BRUSH));
				winclass->lpszClassName = title.getData();
				winclass->hIconSm = LoadIcon(0, IDI_APPLICATION);

				if (RegisterClassEx(winclass) == FALSE)
					throw Util::Exception("Failed to register window class! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}

			handle = CreateWindowEx(0, winclass->lpszClassName, title.getData(), WS_OVERLAPPEDWINDOW,
				CW_USEDEFAULT, CW_USEDEFAULT, size[0], size[1], nullptr, nullptr, winclass->hInstance, this);
			if (handle == nullptr)
				throw Util::Exception("Failed to create window! " + Util::parseWinError(GetLastError(), true), KRY_EXCEPTION_DATA);
			context = GetDC(handle);
			SetForegroundWindow(handle);
			SetFocus(handle);
			destroyed = false;

			PIXELFORMATDESCRIPTOR pfd{0};
			pfd.nSize = sizeof(pfd);
			pfd.nVersion = 1;
			pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_GENERIC_ACCELERATED | PFD_DOUBLEBUFFER;
			pfd.iPixelType = PFD_TYPE_RGBA;
			pfd.cColorBits = 24;
			pfd.cDepthBits = 32;
			pfd.cStencilBits = 8;
			pfd.iLayerType = PFD_MAIN_PLANE;
			int format = ChoosePixelFormat(context, &pfd);
			if (SetPixelFormat(context, format, &pfd) == FALSE)
				throw Util::Exception("Failed to set pixel format! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			glcontext = wglCreateContext(context);
			if (glcontext == nullptr)
				throw Util::Exception("Failed to create render context! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			if (wglMakeCurrent(context, glcontext) == FALSE)
				throw Util::Exception("Failed to switch render context! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			if (glewInit() != GLEW_OK)
				throw Util::Exception("Failed to initialise GLEW! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);

			++wincount; 

			ShowWindow(handle, SW_SHOWNORMAL);
		}

		Window::~Window()
		{
			try
			{
				destroy();
			}
			catch (const Util::Exception& e)
			{
				Util::Logger::log(e);
			}
			
			if (--wincount == 0)
			{ /** #TODO(bug) incorrect MSDN information. unregister class returns 0 no matter what? */
				if (UnregisterClass(winclass->lpszClassName, nullptr) == 0);
					//Util::Logger::log("Failed to unregister window class! " + Util::parseWinError(GetLastError()));
				
				delete winclass;
				winclass = nullptr;
			}
		}

		void Window::setTitle(const Util::String& title)
		{
			if (SetWindowText(handle, title.getData()) == 0)
				throw Util::Exception("Failed to set window title! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}

		void Window::setSize(const Util::Vector2i& size)
		{
			RECT client;
			GetClientRect(handle, &client);
			client.right = client.left + size[0];
			client.bottom = client.top + size[1];

			AdjustWindowRect(&client, static_cast<DWORD>(GetWindowLong(handle, GWL_STYLE)), false);
			SetWindowPos(handle, 0, 0, 0, client.right, client.bottom, SWP_NOMOVE | SWP_NOZORDER);
		}

		void Window::setPosition(const Util::Vector2i& position)
		{
			SetWindowPos(handle, 0, static_cast<int>(position[0]), static_cast<int>(position[1]), 0, 0, SWP_NOSIZE | SWP_NOZORDER);
		}

		Util::String Window::getTitle() const
		{
			int length = GetWindowTextLength(handle);
			char* buffer = new char[length];
			length = GetWindowText(handle, buffer, length);
			if (length == 0)
				throw Util::Exception("Failed to get window title! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			return Util::String(buffer, static_cast<unsigned>(length));
		}

		Util::Vector2i Window::getSize() const
		{
			RECT rect;
			GetClientRect(handle, &rect);
			return Util::Vector2i(rect.right, rect.bottom);
		}

		Util::Vector2i Window::getPosition() const
		{
			RECT rect;
			GetClientRect(handle, &rect);
			POINT client = { rect.right, rect.bottom };
			ClientToScreen(handle, &client);
			return Util::Vector2i(client.x - rect.right, client.y - rect.bottom);
		}

		bool Window::peekEvent(WindowEvent& event)
		{
			event.type = EventType::NONE;

			if (events.empty())
				processEvent();
			if (events.empty())
				return false;

			event = events.front();
			return true;
		}

		bool Window::getEvent(WindowEvent& event)
		{
			event.type = EventType::NONE;

			if (events.empty())
				processEvent();
			if (events.empty())
				return false;

			event = events.front();
			events.pop();
			return true;
		}

		bool Window::waitEvent(WindowEvent& event)
		{
			event.type = EventType::NONE;

			do
				processEvent();
			while (events.empty());

			event = events.front();
			events.pop();
			return true;
		}

		void Window::setMode(WindowMode mode)
		{
			if (winmode == mode)
				return;

			switch (mode)
			{
				case WindowMode::WINDOWED:
				{
					if (SetWindowLong(handle, GWL_STYLE, WS_OVERLAPPEDWINDOW | WS_VISIBLE) == 0)
					{
						throw Util::Exception("Failed to set window mode! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
					}
					SetWindowPos(handle, nullptr, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOMOVE | SWP_NOSIZE);
					UpdateWindow(handle);
					winmode = mode;
				}
					break;
				case WindowMode::BORDERLESS:
				{
					if (SetWindowLong(handle, GWL_STYLE, WS_POPUP | WS_VISIBLE) == 0)
					{
						throw Util::Exception("Failed to set window mode! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
					}
					SetWindowPos(handle, nullptr, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOMOVE | SWP_NOSIZE);
					UpdateWindow(handle);
					winmode = mode;
				}
					break;
				case WindowMode::FULLSCREEN:
				{
					Util::Vector2i screenSize = getScreenSize();

					DEVMODE dvm{0}; // error here is a VS bug (not an actual error)
					dvm.dmSize = sizeof(dvm);
					dvm.dmPelsWidth = screenSize[0];
					dvm.dmPelsHeight = screenSize[1];
					dvm.dmBitsPerPel = 32;
					dvm.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

					if (ChangeDisplaySettings(&dvm, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
					{
						throw Util::Exception("Failed to set window mode! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
					}

					SetWindowPos(handle, HWND_TOPMOST, 0, 0, screenSize[0], screenSize[1], SWP_SHOWWINDOW);
					setMode(WindowMode::BORDERLESS);
					winmode = mode;
				}
					break;
				default:
					throw Util::Exception("Failed to set window mode! Invalid window mode.", KRY_EXCEPTION_DATA);
			}
		}

		void Window::setVsync(bool vsync) const
		{
			wglSwapIntervalEXT((vsync) ? 1 : 0);
		}

		void Window::clipCursor(bool clip) const
		{
			if (clip)
			{
				Util::Vector2i& pos = getPosition();
				Util::Vector2i& size = getSize();
				RECT rect;
				rect.left = pos[0];
				rect.top = pos[1];
				rect.right = pos[0] + size[0];
				rect.bottom = pos[1] + size[1];
				if (ClipCursor(&rect) == FALSE)
					throw Util::Exception("Failed to clip the cursor! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}
			else
			{
				if (ClipCursor(nullptr) == FALSE)
					throw Util::Exception("Failed to un-clip the cursor! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			}
		}

		void Window::hideCursor(bool hide) const
		{
			if (ShowCursor(hide ? TRUE : FALSE) == FALSE)
				throw Util::Exception("Failed to " + (hide ? Util::String("hide") : Util::String("show")) + 
										" the cursor! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
		}

		Util::Vector2i Window::getCursorPos() const
		{
			POINT pos;
			if (GetCursorPos(&pos) == FALSE)
				throw Util::Exception("Failed to get the cursor position! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			if (ScreenToClient(handle, &pos) == FALSE)
				throw Util::Exception("Failed to convert cursor position from screen to client co-ords! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
			return Util::Vector2i(pos.x, pos.y);
		}

		void Window::destroy()
		{
			ShowWindow(handle, SW_MINIMIZE);

			if (glcontext != nullptr)
			{
				wglMakeCurrent(nullptr, nullptr);
				if (wglDeleteContext(glcontext) == FALSE)
				{
					glcontext = nullptr;
					if (context != nullptr)
					{
						ReleaseDC(handle, context); // dont bother with error checking here, dont want to hide the following exception
						context = nullptr;
					}
					if (handle != nullptr)
					{
						DestroyWindow(handle);
						handle = nullptr;
					}
					throw Util::Exception("Failed to destroy render context! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
				}
				glcontext = nullptr;
			}
			if (context != nullptr)
			{
				ReleaseDC(handle, context); /** #TODO(incomplete) add error checking here */
				context = nullptr;
			}
			if (handle != nullptr)
			{
				DestroyWindow(handle);
				handle = nullptr;
			}
			destroyed = true;
		}


		Util::Vector2i Window::getScreenSize(Window *source)
		{
			if (source != nullptr)
			{
				auto monitor = MonitorFromWindow(source->getHandle(), MONITOR_DEFAULTTONEAREST);
				MONITORINFO info{0};
				if (GetMonitorInfo(monitor, &info) == FALSE)
					throw Util::Exception("Failed to retrieve monitor information! " + Util::parseWinError(GetLastError()), KRY_EXCEPTION_DATA);
				return Util::Vector2i(std::abs(info.rcMonitor.right - info.rcMonitor.left), std::abs(info.rcMonitor.bottom - info.rcMonitor.top));
			}
			return Util::Vector2i(GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
		}

		unsigned Window::getMonitorCount()
		{
			return static_cast<unsigned>(GetSystemMetrics(SM_CMONITORS));
		}

		void Window::messageBox(const Util::String& title, const Util::String& message, Window* owner)
		{
			MessageBox(owner != nullptr ? owner->getHandle() : nullptr, message.getData(), title.getData(), MB_OK);
		}
		

		void Window::hookWindowProcess(const WindowProcess& process)
		{
			windowproc = process;
		}

		void Window::processEvent()
		{
			MSG msg;
			while (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}

		long __stdcall Window::winProc(WindowHandle winhandle, unsigned event, unsigned wparam, long lparam)
		{
			if (event == WM_CREATE)
			{
				LONG_PTR win = (LONG_PTR)reinterpret_cast<CREATESTRUCT*>(lparam)->lpCreateParams;
				SetWindowLongPtr(winhandle, GWLP_USERDATA, win);
			}

			Window *window = winhandle != nullptr ? reinterpret_cast<Window*>(GetWindowLongPtr(winhandle, GWLP_USERDATA)) : nullptr;

			if (window != nullptr)
			{
				switch (event)
				{
					case WM_CREATE:
					case WM_CLOSE:
					{
						WindowEvent msg;
						msg.type = EventType::CLOSE;
						msg.open = event == WM_CREATE;
						window->events.push(msg);

						if (event == WM_CLOSE)
							return 0; // dont wanna close the window too early, let the user do it
					}
						break;

					case WM_MOVE:
					{
						WindowEvent msg;
						msg.type = EventType::MOVE;
						msg.position[0] = static_cast<int>(LOWORD(lparam));
						msg.position[1] = static_cast<int>(HIWORD(lparam));
						msg.moving = false;
						window->events.push(msg);
					}
						break;

					case WM_SIZE:
					{
						WindowEvent msg;
						msg.type = EventType::SIZE;
						msg.size[0] = static_cast<int>(LOWORD(lparam));
						msg.size[1] = static_cast<int>(HIWORD(lparam));
						msg.resizing = false;
						window->events.push(msg);
					}
						break;

					case WM_SYSKEYDOWN:
					case WM_KEYDOWN:
					case WM_SYSKEYUP:
					case WM_KEYUP:
					{
						Key key = virtualKeyToKey(wparam, lparam);
						keystates[static_cast<unsigned>(key)] = event == WM_SYSKEYDOWN || event == WM_KEYDOWN;

						WindowEvent msg;
						msg.type = EventType::KEY;
						msg.key = key;
						msg.down = event == WM_SYSKEYDOWN || event == WM_KEYDOWN;

						window->events.push(msg);
					}
						break;

					case WM_LBUTTONDOWN:
					case WM_LBUTTONUP:
					{
						buttonstates[static_cast<unsigned>(Button::LEFT)] = event == WM_LBUTTONDOWN;

						WindowEvent msg;
						msg.type = EventType::MOUSE;
						msg.button = Button::LEFT;
						msg.position[0] = static_cast<int>(LOWORD(lparam));
						msg.position[1] = static_cast<int>(HIWORD(lparam));
						msg.down = event == WM_LBUTTONDOWN;
						window->events.push(msg);
					}
						break;

					case WM_RBUTTONDOWN:
					case WM_RBUTTONUP:
					{
						buttonstates[static_cast<unsigned>(Button::RIGHT)] = event == WM_RBUTTONDOWN;

						WindowEvent msg;
						msg.type = EventType::MOUSE;
						msg.button = Button::RIGHT;
						msg.position[0] = static_cast<int>(LOWORD(lparam));
						msg.position[1] = static_cast<int>(HIWORD(lparam));
						msg.down = event == WM_RBUTTONDOWN;
						window->events.push(msg);
					}
						break;

					case WM_MBUTTONDOWN:
					case WM_MBUTTONUP:
					{
						buttonstates[static_cast<unsigned>(Button::RIGHT)] = event == WM_MBUTTONDOWN;

						WindowEvent msg;
						msg.type = EventType::MOUSE;
						msg.button = Button::MIDDLE;
						msg.position[0] = static_cast<int>(LOWORD(lparam));
						msg.position[1] = static_cast<int>(HIWORD(lparam));
						msg.down = event == WM_MBUTTONDOWN;
						window->events.push(msg);
					}
						break;

					case WM_XBUTTONDOWN:
					case WM_XBUTTONUP:
					{
						Button button = HIWORD(wparam) == XBUTTON1 ? Button::X1 : Button::X2;
						buttonstates[static_cast<unsigned>(button)] = event == WM_XBUTTONDOWN;

						WindowEvent msg;
						msg.type = EventType::MOUSE;
						msg.button = button;
						msg.position[0] = static_cast<int>(LOWORD(lparam));
						msg.position[1] = static_cast<int>(HIWORD(lparam));
						msg.down = event == WM_XBUTTONDOWN;
						window->events.push(msg);
					}
						break;

					case WM_MOUSEWHEEL:
					{
						POINT position;
						position.x = static_cast<short>(LOWORD(lparam));
						position.y = static_cast<short>(HIWORD(lparam));
						ScreenToClient(winhandle, &position);

						WindowEvent msg;
						msg.type = EventType::MOUSE_WHEEL;
						msg.position[0] = position.x;
						msg.position[1] = position.y;
						msg.wheelvalue = static_cast<short>(HIWORD(wparam)) / WHEEL_DELTA;
						window->events.push(msg);
					}
						break;

					case WM_MOUSEMOVE:
					{
						WindowEvent msg;
						msg.type = EventType::MOUSE;
						msg.button = Button::UNKNOWN;
						msg.position[0] = static_cast<int>(LOWORD(lparam));
						msg.position[1] = static_cast<int>(HIWORD(lparam));
						msg.down = false;
						window->events.push(msg);
					}
						break;

					case WM_SETFOCUS:
					case WM_KILLFOCUS:
					{
						WindowEvent msg;
						msg.type = EventType::FOCUS;
						msg.focused = event == WM_SETFOCUS;
						window->events.push(msg);
					}
						break;
				}
				return window->windowproc ? window->windowproc(winhandle, event, wparam, lparam) : DefWindowProc(winhandle, event, wparam, lparam);
			}
			return DefWindowProc(winhandle, event, wparam, lparam);
		}
	}
}
