#include "Utilities\Math.h"
#include <algorithm>

namespace kry
{
	namespace Util
	{
		float radiansToDegrees(const float Radians)
		{
			return Radians * KRY_PIr;
		}

		float degreesToRadians(const float Degrees)
		{
			return Degrees * KRY_PId;
		}

		int clampi(const int Value, const int Min, const int Max)
		{
			return std::max(Min, std::min(Max, Value));
		}

		float clampf(const float Value, const float Min, const float Max)
		{
			return std::max(Min, std::min(Max, Value));
		}

		float clampAngle(const float Value)
		{
			float angle = fmod(Value, 360.0f);
			if (angle < 0.0f) angle += 360.0f;
			return angle;
		}

		bool lineLineInersect(const Vector4f& LineA, const Vector4f& LineB)
		{
			Vector2f p0 = Vector2f(LineA[0], LineA[1]);
			Vector2f p1 = Vector2f(LineA[2], LineA[3]);
			Vector2f p2 = Vector2f(LineB[0], LineB[1]);
			Vector2f p3 = Vector2f(LineB[2], LineB[3]);

			Vector2f const s1 = p1 - p0;
			Vector2f const s2 = p3 - p2;
			Vector2f const u = p0 - p2;
			float const ip = 1.0f / (-s2[0] * s1[1] + s1[0] * s2[1]);
			float const s = (-s1[1] * u[0] + s1[0] * u[1]) * ip;
			float const t = (s2[0] * u[1] - s2[1] * u[0]) * ip;

			return (s >= 0 && s <= 1 && t >= 0 && t <= 1);
		}

		bool boxPointIntersect(const Vector2f& Point, const Vector4f& Diagonals)
		{
			Vector4f diagonals = absBox(Diagonals);

			return (Point[0] >= diagonals[0] && Point[0] <= diagonals[2] && Point[1] >= diagonals[1] && Point[1] <= diagonals[3]);
		}

		bool boxPointIntersect(const Vector2f& Point, const Vector2f& BoxTopLeft, const Vector2f& BoxDimensions)
		{
			return (Point[0] >= BoxTopLeft[0] && Point[0] <= BoxTopLeft[0] + BoxDimensions[0] && Point[1] >= BoxTopLeft[1] && Point[1] <= BoxTopLeft[1] + BoxDimensions[1]);
		}

		bool boxBoxCollision(const Vector4f& DiagonalsA, const Vector4f& DiagonalsB)
		{
			Vector4f diagonalsA = absBox(DiagonalsA);
			Vector4f diagonalsB = absBox(DiagonalsB);

			if (diagonalsA[3] <= diagonalsB[1])
				return false;
			if (diagonalsA[1] >= diagonalsB[3])
				return false;
			if (diagonalsA[2] <= diagonalsB[0])
				return false;
			if (diagonalsA[0] >= diagonalsB[2])
				return false;

			return true;
		}

		bool boxBoxCollision(const Vector2f& BoxTopLeftA, const Vector2f& BoxDimensionsA, const Vector2f& BoxTopLeftB, const Vector2f& BoxDimensionsB)
		{
			if (BoxTopLeftA[1] + BoxDimensionsA[1] <= BoxTopLeftB[1])
				return false;
			if (BoxTopLeftA[1] >= BoxTopLeftB[1] + BoxDimensionsB[1])
				return false;
			if (BoxTopLeftA[0] + BoxDimensionsA[0] <= BoxTopLeftB[0])
				return false;
			if (BoxTopLeftA[0] >= BoxTopLeftB[0] + BoxDimensionsB[0])
				return false;

			return true;
		}

		bool lineBoxIntersect(const Vector4f& Line, const Vector2f& BoxTopLeft, const Vector2f& BoxDimensions)
		{
			Vector2f boxBR = BoxTopLeft + BoxDimensions;

			return (lineLineInersect(Line, Vector4f(BoxTopLeft[0], BoxTopLeft[1], boxBR[0], BoxTopLeft[1]))
				|| lineLineInersect(Line, Vector4f(boxBR[0], BoxTopLeft[1], boxBR[0], boxBR[1]))
				|| lineLineInersect(Line, Vector4f(boxBR[0], boxBR[1], BoxTopLeft[0], boxBR[1]))
				|| lineLineInersect(Line, Vector4f(BoxTopLeft[0], boxBR[1], BoxTopLeft[0], BoxTopLeft[1])));
		}

		bool radialPointIntersect(const Vector2f& Point, const Vector4f& Ellipse)
		{
			Vector2f ellipsePoint = Vector2f(Ellipse[0], Ellipse[1]);
			float radius = ellipseAngleRadius(Vector2f(Ellipse[2], Ellipse[3]), pointDirection(Point, ellipsePoint));

			return (pointDistance(Point, ellipsePoint) <= radius);
		}

		bool radialRadialIntersect(const Vector4f& EllipseA, const Vector4f& EllipseB)
		{
			Vector2f pointA = Vector2f(EllipseA[0], EllipseA[1]);
			Vector2f pointB = Vector2f(EllipseB[0], EllipseB[1]);
			float radiusA = ellipseAngleRadius(Vector2f(EllipseA[2], EllipseA[3]), pointDirection(pointA, pointB));
			float radiusB = ellipseAngleRadius(Vector2f(EllipseB[2], EllipseB[3]), pointDirection(pointB, pointA));

			return (pointDistance(pointA, pointB) <= radiusA + radiusB);
		}

		bool radialBoxIntersect(const Vector4f& Ellipse, const Vector2f& BoxTopLeft, const Vector2f& BoxDimensions)
		{
			Vector2f ellipsePos = Vector2f(Ellipse[0], Ellipse[1]);
			Vector2f ellipseRadius = Vector2f(Ellipse[2], Ellipse[3]);
			Vector2f boxTR = Vector2f(BoxTopLeft[0] + BoxDimensions[0], BoxTopLeft[1]);
			Vector2f boxBR = Vector2f(boxTR[0], BoxTopLeft[1] + BoxDimensions[1]);
			Vector2f boxBL = Vector2f(BoxTopLeft[0], boxBR[1]);

			return (pointDistance(ellipsePos, BoxTopLeft) <= ellipseAngleRadius(ellipseRadius, pointDirection(ellipsePos, BoxTopLeft))
				|| pointDistance(ellipsePos, boxTR) <= ellipseAngleRadius(ellipseRadius, pointDirection(ellipsePos, boxTR))
				|| pointDistance(ellipsePos, boxBR) <= ellipseAngleRadius(ellipseRadius, pointDirection(ellipsePos, boxBR))
				|| pointDistance(ellipsePos, boxBL) <= ellipseAngleRadius(ellipseRadius, pointDirection(ellipsePos, boxBL)));
		}

		float ellipseAngleRadius(const Vector2f& Radius, float Angle)
		{
			Vector2f position;
			float radians = (Angle - 90.0f) * KRY_PId;
			position[0] = Radius[0] * cos(radians);
			position[1] = Radius[1] * sin(radians);

			return pointDistance(Vector2f(), position);
		}

		Vector2f trajectory(const Vector2f& Position, const float Angle, const float Distance)
		{
			Vector2f newPos = Position;
			float radians = (Angle - 90.0f) * KRY_PId;
			newPos[0] += Distance * cos(radians);
			newPos[1] += Distance * sin(radians);

			return newPos;
		}

		Vector3f trajectory(const Vector3f& Position, const Vector3f& Direction, const float Distance)
		{
			return Position + (Direction * Vector3f(Distance));
		}

		float pointDirection(const Vector2f& PointA, const Vector2f& PointB)
		{
			Vector2f difference;
			difference[0] = PointB[0] - PointA[0];
			difference[1] = PointA[1] - PointB[1];
			float angle = atan2(difference[0], difference[1]) * KRY_PIr;

			if (angle < 0)
				angle += 360.0f;

			return angle;
		}

		Vector3f pointDirection(const Vector3f& PointA, const Vector3f& PointB)
		{
			return normalize(PointB - PointA);
		}

		float pointDistance(const Vector2f& PointA, const Vector2f& PointB)
		{
			float x = PointB[0] - PointA[0];
			float y = PointB[1] - PointA[1];

			return sqrt(pow(x, 2) + pow(y, 2));
		}

		float pointDistance(const Vector3f& PointA, const Vector3f& PointB)
		{
			float x = PointB[0] - PointA[0];
			float y = PointB[1] - PointA[1];
			float z = PointB[2] - PointA[2];

			return sqrt(pow(x, 2) + pow(y, 2) + pow(z, 2));
		}

		Vector3f angleToDirection(const Vector3f& Angles)
		{
			float yaw = Angles[1] * KRY_PId;
			float pitch = Angles[0] * KRY_PId;

			return Vector3f(-cos(pitch) * sin(yaw), -sin(pitch), -cos(pitch) * cos(yaw));
		}

		bool angleInCone(const float Angle, const float ConeLeft, const float ConeRight)
		{
			float angle = clampAngle(Angle);
			float coneLeft = clampAngle(ConeLeft);
			float coneRight = clampAngle(ConeRight);
			float range = clampAngle(coneRight - coneLeft);
			float angleRange = clampAngle(coneRight - angle);

			if (range == 0.0f)
				return true;

			return (angleRange >= 0.0f && angleRange <= range);
		}

		Vector2f isoToOrtho(const Vector2f& Coord)
		{
			Vector2f ortho;
			const float radians = atan2(0.5f, 1.0f);
			const float perim = 1.0f / (cos(radians) * 0.5f);
			float x = (Coord[0] / cos(radians)) / perim;
			float y = (Coord[1] / cos(radians)) / perim;
			ortho[0] = x + y;
			ortho[1] = y - (ortho[0] * tan(radians));

			return ortho;
		}

		Vector2f orthoToIso(const Vector2f& Coord)
		{
			Vector2f iso;
			/*float radians = atan2(0.5f, 1.0f);
			float perim = 1.0f / (cos(radians) * 0.5f);
			float x = Coord[0] * sin(radians) * perim;
			float y = Coord[1] * cos(radians) * perim;
			iso[0] = x - y;
			iso[1] = x + y;*/

			float y = Coord[1] * 2.0f;
			iso[0] = Coord[0] - y;
			iso[1] = Coord[0] + y;

			return iso;
		}

		Vector4f absBox(Vector4f Diagonals)
		{
			Vector4f temp = Diagonals;

			if (Diagonals[0] > Diagonals[2] && Diagonals[1] > Diagonals[3])
			{
				Diagonals[0] = temp[2];
				Diagonals[2] = temp[0];
				Diagonals[1] = temp[3];
				Diagonals[3] = temp[1];
			}
			else if (Diagonals[0] < Diagonals[2] && Diagonals[1] > Diagonals[3])
			{
				Diagonals[1] = temp[3];
				Diagonals[3] = temp[1];
			}
			else if (Diagonals[0] > Diagonals[2] && Diagonals[1] < Diagonals[3])
			{
				Diagonals[0] = temp[2];
				Diagonals[2] = temp[0];
			}

			return Diagonals;
		}

		Vector4f boxToDiagonals(const Vector2f& Position, const Vector2f& Dimensions)
		{
			return Vector4f(Position[0], Position[1], Position[0] + Dimensions[0], Position[1] + Dimensions[1]);
		}

		Vector3f normalize(Vector3f Vector)
		{
			float magnitude = sqrtf(pow(Vector[0], 2) + pow(Vector[1], 2) + pow(Vector[2], 2));

			if (magnitude == 0.0f)
				return Vector3f();

			Vector /= Vector3f(magnitude);

			return Vector;
		}

	}
}
